"""Logging configuration for zeek-client."""
import logging

LOG = logging.getLogger(__name__)
LOG.addHandler(logging.NullHandler())


def configure(verbosity=0, rich_logging=False, stream=None):
    """Configures logging.

    Args:
        verbosity (int): the log level. Values 0-3 increase logging, larger
            values make no difference.

        rich_logging (bool): whether to use timestamped, log-style log formatting.
    """
    # Make log levels lower-case, looks better in informal logging
    for level in (logging.ERROR, logging.WARNING, logging.INFO, logging.DEBUG):
        logging.addLevelName(level, logging.getLevelName(level).lower())

    if rich_logging:
        formatter = logging.Formatter(
            "%(asctime)s %(levelname)-8s %(message)s", "%Y-%m-%d %H:%M:%S"
        )
    else:
        formatter = logging.Formatter("%(levelname)s: %(message)s")

    handler = logging.StreamHandler(stream=stream)
    handler.setFormatter(formatter)

    LOG.setLevel(logging.ERROR)

    if verbosity == 1:
        LOG.setLevel(logging.WARNING)
    elif verbosity == 2:
        LOG.setLevel(logging.INFO)
    elif verbosity >= 3:
        LOG.setLevel(logging.DEBUG)

    LOG.addHandler(handler)



# The ZeekControl interactive shell.

from __future__ import print_function
import os
import sys
import logging

from ZeekControl import lock
from ZeekControl import config
from ZeekControl import cmdresult
from ZeekControl import execute
from ZeekControl import control
from ZeekControl import version
from ZeekControl import pluginreg
from ZeekControl import node as node_mod
from ZeekControl.exceptions import *

class TermUI:
    def info(self, txt):
        print(txt)

    def error(self, txt):
        print(txt, file=sys.stderr)
    warn = error


class NullHandler(logging.Handler):
    def emit(self, record):
        pass


def expose(func):
    func.api_exposed = True
    return func

def lock_required(func):
    def wrapper(self, *args, **kwargs):
        self.lock()
        try:
            return func(self, *args, **kwargs)
        finally:
            self.unlock()
    wrapper.lock_required = True
    return wrapper

def lock_required_silent(func):
    def wrapper(self, *args, **kwargs):
        self.lock(showwait=False)
        try:
            return func(self, *args, **kwargs)
        finally:
            self.unlock()
    wrapper.lock_required = True
    return wrapper

def check_config(func):
    def wrapper(self, *args, **kwargs):
        if config.Config.is_cfg_changed():
            self.ui.warn('Configuration has changed. Run the "deploy" command.')
        return func(self, *args, **kwargs)

    return wrapper

class ZeekCtl(object):
    def __init__(self, basedir=version.ZEEKBASE, libdir=version.LIBDIR,
                 libdirinternal=version.LIBDIRINTERNAL, cfgfile=version.CFGFILE,
                 zeekscriptdir=version.ZEEKSCRIPTDIR, ui=TermUI(), state=None):
        self.ui = ui
        self.zeekbase = basedir
        self.libdir = libdir
        self.libdirinternal = libdirinternal

        self.config = config.Configuration(
            self.zeekbase, self.libdir, self.libdirinternal, cfgfile,
            zeekscriptdir, self.ui, state)

        # Remove all log handlers (set by any previous calls to logging.*)
        logging.getLogger().handlers = []

        if self.config.debug:
            # Add a log handler that logs to a file.
            try:
                logging.basicConfig(filename=self.config.debuglog,
                            format="%(asctime)s [%(module)s] %(message)s",
                            datefmt=self.config.timefmt,
                            level=logging.DEBUG)
            except IOError as err:
                raise RuntimeEnvironmentError("%s\nCheck if the user running ZeekControl has write access to the debug log file." % err)
        else:
            # Add a log handler that does nothing.
            h = NullHandler()
            logging.getLogger().addHandler(h)

        self.executor = execute.Executor(self.config)
        self.plugins = pluginreg.PluginRegistry()
        self.setup()
        self.controller = control.Controller(self.config, self.ui, self.executor, self.plugins)

    def setup(self):
        plugindirs = self.config.sitepluginpath.split(":")
        plugindirs.append(self.config.plugindir)
        plugindirs.append(self.config.pluginzeekdir)

        for pdir in plugindirs:
            if pdir:
                self.plugins.addDir(pdir)

        self.plugins.loadPlugins(self.ui, self.executor)
        self.plugins.initPluginOptions()
        self.plugins.addNodeKeys()
        self.config.initPostPlugins()
        self.plugins.initPlugins(self.ui)
        self.plugins.initPluginCmds()
        os.chdir(self.config.zeekbase)
        if self.config.get_state("cronenabled") is None:
            self.config.set_state("cronenabled", True)

    def reload_cfg(self):
        self.config.reload_cfg()

        if self.config.debug:
            if isinstance(logging.getLogger().handlers[0], NullHandler):
                # Remove the null handler and configure logging to a file.
                logging.getLogger().handlers = []
                logging.basicConfig(filename=self.config.debuglog,
                            format="%(asctime)s [%(module)s] %(message)s",
                            datefmt=self.config.timefmt,
                            level=logging.DEBUG)

            # Re-enable all log levels.
            logging.disable(logging.NOTSET)
        else:
            # Disable logging to all log levels that we use.
            logging.disable(logging.CRITICAL)

        self.executor.finish()
        self.plugins.initPluginOptions()
        self.config.initPostPlugins()
        self.plugins.initPlugins(self.ui)
        self.plugins.initPluginCmds()

    def finish(self):
        self.executor.finish()
        self.plugins.finishPlugins()

    def warn_zeekctl_install(self):
        self.config.warn_zeekctl_install()

    # Turns node name arguments into a list of nodes.  If "get_hosts" is True,
    # then only one node per host is chosen.  If "get_types" is True, then
    # only one node per node type (manager, proxy, etc.) is chosen.
    def node_args(self, args=None, get_hosts=False, get_types=False):
        nodes = []

        if args:
            for arg in args.split():
                nodelist = self.config.nodes(arg)
                if not nodelist:
                    raise InvalidNodeError("unknown node '%s'" % arg)

                nodes += nodelist

            # Remove duplicate nodes
            newlist = list(set(nodes))
            if len(newlist) != len(nodes):
                nodes = newlist
        else:
            # Get all nodes.
            nodes = self.config.nodes()

        # Sort the list so that it doesn't depend on initial order of arguments
        nodes.sort(key=node_mod.sortnode)

        if get_hosts:
            hosts = {}
            hostnodes = []
            for node in nodes:
                if node.host not in hosts:
                    hosts[node.host] = 1
                    hostnodes.append(node)

            nodes = hostnodes

        if get_types:
            types = {}
            typenodes = []
            for node in nodes:
                if node.type not in types:
                    types[node.type] = 1
                    typenodes.append(node)

            nodes = typenodes

        return nodes

    def lock(self, showwait=True):
        lockstatus = lock.lock(self.ui, showwait)
        if not lockstatus:
            raise LockError("Unable to get lock")

        self.config.read_state()

    def unlock(self):
        lock.unlock(self.ui)

    def node_names(self):
        return [ n.name for n in self.config.nodes() ]

    def node_groups(self):
        return node_mod.node_groups()

    @expose
    @check_config
    def nodes(self):
        results = cmdresult.CmdResult()

        if self.plugins.cmdPre("nodes"):
            for n in self.config.nodes():
                results.set_node_data(n, True, n.to_dict())
        else:
            results.ok = False

        self.plugins.cmdPost("nodes")

        return results

    @expose
    @check_config
    def get_config(self):
        results = cmdresult.CmdResult()

        if self.plugins.cmdPre("config"):
            results.keyval = self.config.options()
        else:
            results.ok = False

        self.plugins.cmdPost("config")
        return results

    @expose
    @check_config
    @lock_required
    def install(self, local=False):
        if self.plugins.cmdPre("install"):
            results = self.controller.install(local)
        else:
            results = cmdresult.CmdResult(ok=False)

        self.plugins.cmdPost("install")
        return results

    @expose
    @check_config
    @lock_required
    def start(self, node_list=None):
        nodes = self.node_args(node_list)

        nodes = self.plugins.cmdPreWithNodes("start", nodes)
        results = self.controller.start(nodes)
        self.plugins.cmdPostWithResults("start", results.get_node_data())

        return results

    @expose
    @check_config
    @lock_required
    def stop(self, node_list=None):
        nodes = self.node_args(node_list)

        nodes = self.plugins.cmdPreWithNodes("stop", nodes)
        results = self.controller.stop(nodes)
        self.plugins.cmdPostWithResults("stop", results.get_node_data())

        return results

    @expose
    @check_config
    @lock_required
    def restart(self, clean=False, node_list=None):
        nodes = self.node_args(node_list)

        nodes = self.plugins.cmdPreWithNodes("restart", nodes, clean)

        self.ui.info("stopping ...")
        results = self.stop(node_list)
        if not results.ok:
            return results

        if clean:
            self.ui.info("cleaning up ...")
            results = self.cleanup(node_list=node_list)
            if not results.ok:
                return results

            self.ui.info("checking configurations ...")
            results = self.check(node_list)
            if not results.ok:
                return results

            self.ui.info("installing ...")
            results = self.install()
            if not results.ok:
                return results

        self.ui.info("starting ...")
        results = self.start(node_list)

        self.plugins.cmdPostWithNodes("restart", nodes)
        return results

    @expose
    @lock_required
    def deploy(self):
        if not self.plugins.cmdPre("deploy"):
            results = cmdresult.CmdResult(ok=False)
            return results

        if self.config.is_cfg_changed():
            self.ui.info("Reloading zeekctl configuration ...")
            self.reload_cfg()

        self.ui.info("checking configurations ...")
        results = self.check(check_node_types=True)
        if not results.ok:
            for (node, success, output) in results.get_node_output():
                if not success:
                    self.ui.info("%s scripts failed." % node)
                    self.ui.info(output)

            return results

        self.ui.info("installing ...")
        results = self.install()
        if not results.ok:
            return results

        self.ui.info("stopping ...")
        results = self.stop()
        if not results.ok:
            return results

        self.ui.info("starting ...")
        results = self.start()

        self.plugins.cmdPost("deploy")
        return results

    @expose
    @check_config
    @lock_required
    def status(self, node_list=None):
        nodes = self.node_args(node_list)

        nodes = self.plugins.cmdPreWithNodes("status", nodes)
        results = self.controller.status(nodes)
        self.plugins.cmdPostWithNodes("status", nodes)
        return results

    @expose
    @lock_required
    def top(self, node_list=None):
        nodes = self.node_args(node_list)

        nodes = self.plugins.cmdPreWithNodes("top", nodes)
        results = self.controller.top(nodes)
        self.plugins.cmdPostWithNodes("top", nodes)

        return results

    @expose
    @check_config
    @lock_required
    def diag(self, node_list=None):
        nodes = self.node_args(node_list)

        nodes = self.plugins.cmdPreWithNodes("diag", nodes)
        results = self.controller.diag(nodes)
        self.plugins.cmdPostWithNodes("diag", nodes)

        return results

    @expose
    @lock_required_silent
    def cron(self, watch=True):
        if self.plugins.cmdPre("cron", "", watch):
            self.controller.cron(watch)
        self.plugins.cmdPost("cron", "", watch)

        return True

    @expose
    @check_config
    @lock_required
    def cronenabled(self):
        results = False
        if self.plugins.cmdPre("cron", "?", False):
            results = self.config.cronenabled
        self.plugins.cmdPost("cron", "?", False)
        return results

    @expose
    @check_config
    @lock_required
    def setcronenabled(self, enable=True):
        if enable:
            if self.plugins.cmdPre("cron", "enable", False):
                self.config.set_state("cronenabled", True)
                self.ui.info("cron enabled")
            self.plugins.cmdPost("cron", "enable", False)
        else:
            if self.plugins.cmdPre("cron", "disable", False):
                self.config.set_state("cronenabled", False)
                self.ui.info("cron disabled")
            self.plugins.cmdPost("cron", "disable", False)

        return True

    @expose
    @check_config
    @lock_required
    def check(self, node_list=None, check_node_types=False):
        nodes = self.node_args(node_list, get_types=check_node_types)

        nodes = self.plugins.cmdPreWithNodes("check", nodes)
        results = self.controller.check(nodes)
        self.plugins.cmdPostWithResults("check", results.get_node_data())

        return results

    @expose
    @check_config
    @lock_required
    def cleanup(self, cleantmp=False, node_list=None):
        nodes = self.node_args(node_list)

        nodes = self.plugins.cmdPreWithNodes("cleanup", nodes, cleantmp)
        results = self.controller.cleanup(nodes, cleantmp)
        self.plugins.cmdPostWithNodes("cleanup", nodes, cleantmp)

        return results

    @expose
    @check_config
    @lock_required
    def capstats(self, interval=10, node_list=None):
        nodes = self.node_args(node_list)
        nodes = self.plugins.cmdPreWithNodes("capstats", nodes, interval)
        results = self.controller.capstats(nodes, interval)
        self.plugins.cmdPostWithNodes("capstats", nodes, interval)

        return results

    @expose
    @check_config
    @lock_required
    def df(self, node_list=None):
        nodes = self.node_args(node_list, get_hosts=True)
        nodes = self.plugins.cmdPreWithNodes("df", nodes)
        results = self.controller.df(nodes)
        self.plugins.cmdPostWithNodes("df", nodes)

        return results

    @expose
    @check_config
    @lock_required
    def print_id(self, id, node_list=None):
        nodes = self.node_args(node_list)
        nodes = self.plugins.cmdPreWithNodes("print", nodes, id)
        results = self.controller.print_id(nodes, id)
        self.plugins.cmdPostWithNodes("print", nodes, id)

        return results

    @expose
    @check_config
    @lock_required
    def peerstatus(self, node_list=None):
        nodes = self.node_args(node_list)
        nodes = self.plugins.cmdPreWithNodes("peerstatus", nodes)
        results = self.controller.peerstatus(nodes)
        self.plugins.cmdPostWithNodes("peerstatus", nodes)

        return results

    @expose
    @check_config
    @lock_required
    def netstats(self, node_list=None):
        if not node_list:
            node_list = None
            if not self.config.standalone:
                node_list = node_mod.worker_group()

        nodes = self.node_args(node_list)
        nodes = self.plugins.cmdPreWithNodes("netstats", nodes)
        results = self.controller.netstats(nodes)
        self.plugins.cmdPostWithNodes("netstats", nodes)

        return results

    @expose
    @check_config
    def execute(self, cmd):
        nodes = self.node_args(get_hosts=True)

        if self.plugins.cmdPre("exec", cmd):
            results = self.controller.execute_cmd(nodes, cmd)
        else:
            results = cmdresult.CmdResult(ok=False)

        self.plugins.cmdPost("exec", cmd)

        return results

    @expose
    @check_config
    @lock_required
    def scripts(self, check=False, node_list=None):
        nodes = self.node_args(node_list)

        nodes = self.plugins.cmdPreWithNodes("scripts", nodes, check)
        results = self.controller.scripts(nodes, check)
        self.plugins.cmdPostWithNodes("scripts", nodes, check)

        return results

    @expose
    @check_config
    @lock_required
    def process(self, trace, options, scripts):
        if self.plugins.cmdPre("process", trace, options, scripts):
            results = self.controller.process(trace, options, scripts)
        else:
            results = cmdresult.CmdResult(ok=False)

        self.plugins.cmdPost("process", trace, options, scripts, results.ok)

        return results

    @expose
    @check_config
    @lock_required
    def plugincmd(self, cmd, args):
        return self.plugins.runCustomCommand(cmd, args, self.ui)




#
# ZeekControl Plugin API.
#

from __future__ import print_function
import logging

from ZeekControl import config
from ZeekControl import doc

class Plugin(object):
    """The class ``Plugin`` is the base class for all ZeekControl plugins.

    The class has a number of methods for plugins to override, and every
    plugin must at least override ``name()`` and ``pluginVersion()``.

    For each ZeekControl command ``foo``, there are two methods,
    ``cmd_foo_pre`` and ``cmd_foo_post``, that are called just before the
    command is executed and just after it has finished, respectively. The
    arguments these methods receive correspond to their command-line
    parameters, and are further documented below.

    The ``cmd_<XXX>_pre`` methods have the ability to prevent the command's
    execution, either completely or partially for those commands that take
    nodes as parameters. In the latter case, the method receives a list of
    nodes that the command is to be run on, and it can filter that list and
    returns modified version of nodes to actually use. The standard case would
    be returning simply the unmodified ``nodes`` parameter. To completely
    block the command's execution, return an empty list. To just not execute
    the command for a subset, remove the affected ones.  For commands that do
    not receive nodes as arguments, the return value is interpreted as boolean
    indicating whether command execution should proceed (True) or not (False).

    The ``cmd_<XXX>_post`` methods likewise receive the commands arguments as
    their parameter, as documented below. For commands taking nodes, the list
    corresponds to those nodes for which the command was actually executed
    (i.e., after any ``cmd_<XXX>_pre`` filtering).

    Note that if a plugin prevents a command from executing either completely or
    partially, it should report its reason via the ``message()`` or
    ``error()`` methods.

    If multiple plugins hook into the same command, all their
    ``cmd_<XXX>_{pre,post}`` are executed in undefined order. The command is
    executed on the intersection of all ``cmd_<XXX>_pre`` results.

    Finally, note that the ``restart`` command is just a combination of other
    commands and thus their callbacks are run in addition to the callbacks
    for ``restart``.
    """

    def __init__(self, apiversion):
        """Must be called by the plugin with the plugin API version it
        expects to use. The version currently documented here is 1."""
        self._apiversion = apiversion
        self.activated = False

    def apiVersion(self):
        """Returns the plugin API that the plugin expects to use."""
        return self._apiversion

    @doc.api
    def getGlobalOption(self, name):
        """Returns the value of the global ZeekControl option *name*.

        See the output of ``zeekctl config`` for a complete list."""
        val = config.Config.get_option(name)
        if val is None:
            raise KeyError("plugin %s lookup of unknown config option %s" % (self.name(), name))

        return val

    @doc.api
    def getOption(self, name):
        """Returns the value of one of the plugin's options, *name*.

        An option has a default value (see *options()*), which can be
        overridden by a user in ``zeekctl.cfg``. An option's value cannot be
        changed by the plugin.
        """
        name = "%s.%s" % (self.prefix(), name)

        val = config.Config.get_option(name)
        if val is None:
            raise KeyError("plugin %s lookup of unknown plugin option %s" % (self.name(), name))

        return val

    @doc.api
    def getState(self, name):
        """Returns the current value of one of the plugin's state variables,
        *name*. If it has not yet been set, an empty string will be returned.

        Different from options, state variables can be set by the plugin.
        They are persistent across restarts.

        Note that a plugin cannot query any global ZeekControl state variables.
        """
        name = "%s.state.%s" % (self.prefix(), name)

        return config.Config.get_state(name, "")

    @doc.api
    def setState(self, name, value):
        """Sets one of the plugin's state variables, *name*, to *value*.
        The change is permanent and will be recorded to disk.

        Note that a plugin cannot change any global ZeekControl state
        variables.
        """
        if "." in name or " " in name:
            self.error('plugin %s state variable name "%s" must not contain dots or spaces' % (self.name(), name))
            return

        name = "%s.state.%s" % (self.prefix(), name)
        config.Config.set_state(name, value)

    @doc.api
    def parseNodes(self, names):
        """Returns a tuple which contains two lists. The first list is a list
        of `Node`_ objects for a string of space-separated node names. If a
        name does not correspond to a known node, then the name is added
        to the second list in the returned tuple.
        """
        nodes = []
        notnodes = []

        for arg in names.split():
            nodelist = config.Config.nodes(arg)
            if nodelist:
                nodes += nodelist
            else:
                notnodes.append(arg)

        # Sort the list so that it doesn't depend on initial order of arguments
        nodes.sort(key=lambda n: (n.type, n.name))

        return (nodes, notnodes)

    @doc.api
    def message(self, msg):
        """Reports a message to the user."""
        print("%s" % msg)

    @doc.api
    def debug(self, msg):
        """Logs a debug message in ZeekControl's debug log if enabled."""
        logging.debug("%s: %s", self.prefix(), msg)

    @doc.api
    def error(self, msg):
        """Reports an error to the user."""
        print("error: %s" % msg)

    @doc.api
    def execute(self, node, cmd):
        """Executes a command on the host for the given *node* of type
        `Node`_. Returns a tuple ``(success, output)`` in which ``success`` is
        True if the command ran successfully, and ``output`` is a string
        which contains the combined stdout/stderr output."""

        resultlist = self.executor.run_shell_cmds([(node, cmd)])
        if resultlist:
            _, success, output = resultlist[0]
        else:
            success = False
            output = ""

        return (success, output)

    @doc.api
    def nodes(self):
        """Returns a list of all configured `Node`_ objects."""
        return config.Config.nodes()

    @doc.api
    def hosts(self, nodes=[]):
        """Returns a list of Node_ objects which is a subset of the list in
        *nodes*, such that only one node per host will be chosen.  If *nodes*
        is empty, then the returned list will be a subset of the entire list
        of configured nodes."""

        if not nodes:
            return [n for n in config.Config.hosts()]

        result = []
        h = {}

        for n in nodes:
            if n.host not in h:
                h[n.host] = 1
                result.append(n)

        return result

    @doc.api
    def executeParallel(self, cmds):
        """Executes a set of commands in parallel on multiple hosts. ``cmds``
        is a list of tuples ``(node, cmd)``, in which the *node* is a `Node`_
        instance and *cmd* is a string with the command to execute for it. The
        method returns a list of tuples ``(node, success, output)``, in which
        ``success`` is True if the command ran successfully, and ``output`` is
        a string containing the combined stdout/stderr output for the
        corresponding ``node``."""

        return self.executor.run_shell_cmds(cmds)

    ### Methods that must be overridden by plugins.

    @doc.api("override")
    def name(self):
        """Returns a string with a descriptive name for the plugin (e.g.,
        ``"TestPlugin"``). The name must not contain any whitespace.

        This method must be overridden by derived classes. The implementation
        must not call the parent class' implementation.
        """
        raise NotImplementedError

    @doc.api("override")
    def pluginVersion(self):
        """
        Returns an integer with a version number for the plugin. Plugins
        should increase their version number with any significant change.

        This method must be overridden by derived classes. The implementation
        must not call the parent class' implementation.
        """
        raise NotImplementedError

    ### Methods that can be overridden by plugins.

    @doc.api("override")
    def prefix(self):
        """Returns a string with a prefix for the plugin's options and
        commands names (e.g., "myplugin").  The prefix cannot contain
        any whitespace or dots (because dots are used as separators when
        forming the plugin's option names, state variable names, and
        command names).

        Note that ZeekControl will refuse to load a plugin if its prefix
        matches the prefix of another loaded plugin (this comparison is not
        case-sensitive).

        This method can be overridden by derived classes. The implementation
        must not call the parent class' implementation. The default
        implementation returns a lower-cased version of *name()*.
        """
        return self.name().lower()

    @doc.api("override")
    def options(self):
        """Returns a set of local configuration options provided by the
        plugin.

        The return value is a list of 4-tuples each having the following
        elements:

            ``name``
                A string with name of the option (e.g., ``Path``). Option
                names are not case-sensitive. Note that the option name exposed
                to the user will be prefixed with your plugin's prefix as
                returned by *prefix()* (e.g., ``myplugin.Path``).

            ``type``
                A string with type of the option, which must be one of
                ``"bool"``, ``"string"``, or ``"int"``.

            ``default``
                The option's default value.  Note that this value must be
                enclosed in quotes if the type is "string", and must not be
                enclosed in quotes if the type is not "string".

            ``description``
                A string with a description of the option semantics.

        This method can be overridden by derived classes. The implementation
        must not call the parent class' implementation. The default
        implementation returns an empty list.
        """
        return []

    @doc.api("override")
    def commands(self):
        """Returns a set of custom commands provided by the
        plugin.

        The return value is a list of 3-tuples each having the following
        elements:

            ``command``
                A string with the command's name. Note that the command name
                exposed to the user will be prefixed with the plugin's prefix
                as returned by *prefix()* (e.g., ``myplugin.mycommand``, or
                just ``myplugin`` if the command name is an empty string).

            ``arguments``
                A string describing the command's arguments in a textual form
                suitable for use in the ``help`` command summary (e.g.,
                ``[<nodes>]`` for a command taking an optional list of nodes).
                Empty if no arguments are expected.

            ``description``
                A string with a description of the command's semantics suitable
                for use in the ``help`` command summary.


        This method can be overridden by derived classes. The implementation
        must not call the parent class' implementation. The default
        implementation returns an empty list.
        """
        return []

    @doc.api("override")
    def nodeKeys(self):
        """Returns a list of names of custom keys for nodes (the value of a
        key can be specified in ``node.cfg`` for any node defined there).
        Node key names are not case-sensitive.

        The value for a key will be available from the `Node`_ object as
        attribute ``<prefix>_<key>`` (e.g., ``node.myplugin_mykey``). If not
        set, the attribute will be set to an empty string.

        This method can be overridden by derived classes. The implementation
        must not call the parent class' implementation. The default
        implementation returns an empty list.
        """
        return []

    @doc.api("override")
    def zeekctl_config(self):
        """Returns a string containing Zeek script code that should be written
        to the dynamically generated Zeek script named "zeekctl-config.zeek".
        This provides a way for plugins to easily add Zeek script code that
        depends on zeekctl settings.

        This method can be overridden by derived classes. The default
        implementation does nothing.
        """
        return

    def broctl_config(self):
        """Deprecated legacy name for `zeekctl_config`."""
        return

    @doc.api("override")
    def init(self):
        """Called once just before ZeekControl starts executing any commands.
        This method can do any initialization that the plugin may require.

        Note that when this method executes, ZeekControl guarantees that all
        internals are fully set up (e.g., user-defined options are available).
        This may not be the case when the class ``__init__`` method runs.

        Returns a boolean, indicating whether the plugin should be used. If it
        returns ``False``, the plugin will be removed and no other methods
        called.

        This method can be overridden by derived classes. The default
        implementation always returns True.
        """
        return True

    @doc.api("override")
    def done(self):
        """Called once just before ZeekControl terminates. This method can do
        any cleanup the plugin may require.

        This method can be overridden by derived classes. The default
        implementation does nothing.
        """
        return

    @doc.api("override")
    def hostStatusChanged(self, host, status):
        """Called when ZeekControl's ``cron`` command finds the availability of
        a cluster system to have changed. Initially, all systems are assumed
        to be up and running. Once ZeekControl notices that a system isn't
        responding (defined as not accepting SSH sessions), it calls
        this method, passing in a string with
        the name of the *host* and a boolean *status* set to False. Once the
        host becomes available again, the method will be called again for the
        same host with *status* now set to True.

        Note that ZeekControl's ``cron`` tracks a host's availability across
        execution, so if the next time it's run the host is still down, this
        method will not be called again.

        This method can be overridden by derived classes. The default
        implementation does nothing.
        """
        return

    @doc.api("override")
    def zeekProcessDied(self, node):
        """Called when ZeekControl finds the Zeek process for Node_ *node*
        to have terminated unexpectedly. This method will be called just
        before ZeekControl prepares the node's "crash report" and before it
        cleans up the node's spool directory.

        This method can be overridden by derived classes. The default
        implementation does nothing.
        """
        return

    def broProcessDied(self, node):
        """Deprecated legacy name for `zeekProcessDied`."""
        # We keep this around as it's difficult to warn about its usage.
        return

    # Per-command help currently not supported by zeekctl. May add this later.
    #
    #@doc.api(override):
    #def help_custom(self, cmd):
    #    """Called for getting the ``help`` text for a custom command defined
    #    by Plugin.commands_. Returns a string with the text, or an empty
    #    string if no help is available.
    #
    #    This method can be overridden by derived classes. The default
    #    implementation always returns an empty string.
    #    """
    #    return ""

    @doc.api("override")
    def cmd_nodes_pre(self):
        """Called just before the ``nodes`` command is run. Returns a
        boolean indicating whether or not the command should run.

        This method can be overridden by derived classes. The default
        implementation does nothing.
        """
        return True

    @doc.api("override")
    def cmd_nodes_post(self):
        """Called just after the ``nodes`` command has finished.

        This method can be overridden by derived classes. The default
        implementation does nothing.
        """
        pass

    @doc.api("override")
    def cmd_config_pre(self):
        """Called just before the ``config`` command is run. Returns a boolean
        indicating whether or not the command should run.

        This method can be overridden by derived classes. The default
        implementation does nothing.
        """
        return True

    @doc.api("override")
    def cmd_config_post(self):
        """Called just after the ``config`` command has finished.

        This method can be overridden by derived classes. The default
        implementation does nothing.
        """
        pass

    @doc.api("override")
    def cmd_exec_pre(self, cmdline):
        """Called just before the ``exec`` command is run. *cmdline* is a
        string with the command line to execute.

        Returns a boolean indicating whether or not the ``exec`` command
        should run.

        This method can be overridden by derived classes. The default
        implementation does nothing.
        """
        return True

    @doc.api("override")
    def cmd_exec_post(self, cmdline):
        """Called just after the ``exec`` command has finished. Arguments are
        as with the ``pre`` method.

        This method can be overridden by derived classes. The default
        implementation does nothing.
        """
        pass

    @doc.api("override")
    def cmd_install_pre(self):
        """Called just before the ``install`` command is run. Returns a
        boolean indicating whether or not the command should run.

        This method can be overridden by derived classes. The default
        implementation does nothing.
        """
        return True

    @doc.api("override")
    def cmd_install_post(self):
        """Called just after the ``install`` command has finished.

        This method can be overridden by derived classes. The default
        implementation does nothing.
        """
        pass

    @doc.api("override")
    def cmd_cron_pre(self, arg, watch):
        """Called just before the ``cron`` command is run. *arg* is an empty
        string if the command is executed without arguments. Otherwise, it is
        one of the strings: ``enable``, ``disable``, ``?``. *watch* is a
        boolean indicating whether the ``cron`` command should restart
        abnormally terminated Zeek processes; it's only valid if *arg* is empty.

        Returns a boolean indicating whether or not the ``cron`` command should
        run.

        This method can be overridden by derived classes. The default
        implementation does nothing.
        """
        return True

    @doc.api("override")
    def cmd_cron_post(self, arg, watch):
        """Called just after the ``cron`` command has finished. Arguments are
        as with the ``pre`` method.

        This method can be overridden by derived classes. The default
        implementation does nothing.
        """
        pass

    @doc.api("override")
    def cmd_check_pre(self, nodes):
        """Called just before the ``check`` command is run. It receives the
        list of nodes, and returns the list of nodes that should proceed with
        the command.

        This method can be overridden by derived classes. The default
        implementation does nothing.
        """
        pass

    @doc.api("override")
    def cmd_check_post(self, results):
        """Called just after the ``check`` command has finished. It receives
        the list of 2-tuples ``(node, bool)`` indicating the nodes the command
        was executed for, along with their success status.

        This method can be overridden by derived classes. The default
        implementation does nothing.
        """
        pass

    @doc.api("override")
    def cmd_start_pre(self, nodes):
        """Called just before the ``start`` command is run. It receives the
        list of nodes, and returns the list of nodes that should proceed with
        the command.

        This method can be overridden by derived classes. The default
        implementation does nothing.
        """
        pass

    @doc.api("override")
    def cmd_start_post(self, results):
        """Called just after the ``start`` command has finished. It receives
        the list of 2-tuples ``(node, bool)`` indicating the nodes the command
        was executed for, along with their success status.

        This method can be overridden by derived classes. The default
        implementation does nothing.
        """
        pass

    @doc.api("override")
    def cmd_stop_pre(self, nodes):
        """Called just before the ``stop`` command is run. It receives the
        list of nodes, and returns the list of nodes that should proceed with
        the command.

        This method can be overridden by derived classes. The default
        implementation does nothing.
        """
        pass

    @doc.api("override")
    def cmd_stop_post(self, results):
        """Called just after the ``stop`` command has finished. It receives
        the list of 2-tuples ``(node, bool)`` indicating the nodes the command
        was executed for, along with their success status.

        This method can be overridden by derived classes. The default
        implementation does nothing.
        """
        pass

    @doc.api("override")
    def cmd_deploy_pre(self):
        """Called just before the ``deploy`` command is run. Returns a
        boolean indicating whether or not the command should run.

        This method can be overridden by derived classes. The default
        implementation does nothing.
        """
        return True

    @doc.api("override")
    def cmd_deploy_post(self):
        """Called just after the ``deploy`` command has finished.

        This method can be overridden by derived classes. The default
        implementation does nothing.
        """
        pass

    @doc.api("override")
    def cmd_status_pre(self, nodes):
        """Called just before the ``status`` command is run. It receives the
        list of nodes, and returns the list of nodes that should proceed with
        the command.

        This method can be overridden by derived classes. The default
        implementation does nothing.
        """
        pass

    @doc.api("override")
    def cmd_status_post(self, nodes):
        """Called just after the ``status`` command has finished.  Arguments
        are as with the ``pre`` method.

        This method can be overridden by derived classes. The default
        implementation does nothing.
        """
        pass

    @doc.api("override")
    def cmd_update_pre(self, nodes):
        """Called just before the ``update`` command is run. It receives the
        list of nodes, and returns the list of nodes that should proceed with
        the command.

        This method can be overridden by derived classes. The default
        implementation does nothing.
        """
        pass

    @doc.api("override")
    def cmd_update_post(self, results):
        """Called just after the ``update`` command has finished. It receives
        the list of 2-tuples ``(node, bool)`` indicating the nodes the command
        was executed for, along with their success status.

        This method can be overridden by derived classes. The default
        implementation does nothing.
        """
        pass

    @doc.api("override")
    def cmd_custom(self, cmd, args, cmdout):
        """Called when a command defined by the ``commands`` method is executed.
        *cmd* is the command (without the plugin's prefix), and *args* is a
        single string with all arguments.  It returns a CmdResult object
        containing the command results.

        If the arguments are actually node names, ``parseNodes`` can
        be used to get the `Node`_ objects.

        This method can be overridden by derived classes. The default
        implementation does nothing.
        """
        pass

    @doc.api("override")
    def cmd_df_pre(self, nodes):
        """Called just before the ``df`` command is run. It receives the
        list of nodes, and returns the list of nodes that should proceed with
        the command.

        This method can be overridden by derived classes. The default
        implementation does nothing.
        """
        pass

    @doc.api("override")
    def cmd_df_post(self, nodes):
        """Called just after the ``df`` command has finished. Arguments are as
        with the ``pre`` method.

        This method can be overridden by derived classes. The default
        implementation does nothing.
        """
        pass

    @doc.api("override")
    def cmd_diag_pre(self, nodes):
        """Called just before the ``diag`` command is run. It receives the
        list of nodes, and returns the list of nodes that should proceed with
        the command.

        This method can be overridden by derived classes. The default
        implementation does nothing.
        """
        pass

    @doc.api("override")
    def cmd_diag_post(self, nodes):
        """Called just after the ``diag`` command has finished. Arguments are
        as with the ``pre`` method.

        This method can be overridden by derived classes. The default
        implementation does nothing.
        """
        pass

    @doc.api("override")
    def cmd_peerstatus_pre(self, nodes):
        """Called just before the ``peerstatus`` command is run. It receives the
        list of nodes, and returns the list of nodes that should proceed with
        the command.

        This method can be overridden by derived classes. The default
        implementation does nothing.
        """
        pass

    @doc.api("override")
    def cmd_peerstatus_post(self, nodes):
        """Called just after the ``peerstatus`` command has finished.
        Arguments are as with the ``pre`` method.

        This method can be overridden by derived classes. The default
        implementation does nothing.
        """
        pass

    @doc.api("override")
    def cmd_netstats_pre(self, nodes):
        """Called just before the ``netstats`` command is run. It receives the
        list of nodes, and returns the list of nodes that should proceed with
        the command.

        This method can be overridden by derived classes. The default
        implementation does nothing.
        """
        pass

    @doc.api("override")
    def cmd_netstats_post(self, nodes):
        """Called just after the ``netstats`` command has finished. Arguments
        are as with the ``pre`` method.

        This method can be overridden by derived classes. The default
        implementation does nothing.
        """
        pass

    @doc.api("override")
    def cmd_top_pre(self, nodes):
        """Called just before the ``top`` command is run. It receives the list
        of nodes, and returns the list of nodes that should proceed with the
        command. Note that when ``top`` is run interactively to auto-refresh
        continuously, this method will be called once before each update.

        This method can be overridden by derived classes. The default
        implementation does nothing.
        """
        pass

    @doc.api("override")
    def cmd_top_post(self, nodes):
        """Called just after the ``top`` command has finished. Arguments are
        as with the ``pre`` method. Note that when ``top`` is run
        interactively to auto-refresh continuously, this method will be called
        once after each update.

        This method can be overridden by derived classes. The default
        implementation does nothing.
        """
        pass

    @doc.api("override")
    def cmd_restart_pre(self, nodes, clean):
        """Called just before the ``restart`` command is run. It receives the
        list of nodes, and returns the list of nodes that should proceed with
        the command. *clean* is boolean indicating whether the ``--clean``
        argument has been given.

        This method can be overridden by derived classes. The default
        implementation does nothing.
        """
        pass

    @doc.api("override")
    def cmd_restart_post(self, nodes):
        """Called just after the ``restart`` command has finished. It receives
        a list of *nodes* indicating the nodes on which the command was
        executed.

        This method can be overridden by derived classes. The default
        implementation does nothing.
        """
        pass

    @doc.api("override")
    def cmd_cleanup_pre(self, nodes, all):
        """Called just before the ``cleanup`` command is run. It receives the
        list of nodes, and returns the list of nodes that should proceed with
        the command. *all* is boolean indicating whether the ``--all``
        argument has been given.

        This method can be overridden by derived classes. The default
        implementation does nothing.
        """
        pass

    @doc.api("override")
    def cmd_cleanup_post(self, nodes, all):
        """Called just after the ``cleanup`` command has finished. Arguments
        are as with the ``pre`` method.

        This method can be overridden by derived classes. The default
        implementation does nothing.
        """
        pass

    @doc.api("override")
    def cmd_capstats_pre(self, nodes, interval):
        """Called just before the ``capstats`` command is run. It receives the
        list of nodes, and returns the list of nodes that should proceed with
        the command. *interval* is an integer with the measurement interval in
        seconds.

        This method can be overridden by derived classes. The default
        implementation does nothing.
        """
        pass

    @doc.api("override")
    def cmd_capstats_post(self, nodes, interval):
        """Called just after the ``capstats`` command has finished. Arguments
        are as with the ``pre`` method.

        This method can be overridden by derived classes. The default
        implementation does nothing.
        """
        pass

    @doc.api("override")
    def cmd_scripts_pre(self, nodes, check):
        """Called just before the ``scripts`` command is run. It receives the
        list of nodes, and returns the list of nodes that should proceed with
        the command. *check* is boolean indicating whether the ``-c``
        option was given.

        This method can be overridden by derived classes. The default
        implementation does nothing.
        """
        pass

    @doc.api("override")
    def cmd_scripts_post(self, nodes, check):
        """Called just after the ``scripts`` command has finished. Arguments
        are as with the ``pre`` method.

        This method can be overridden by derived classes. The default
        implementation does nothing.
        """
        pass

    @doc.api("override")
    def cmd_print_pre(self, nodes, id):
        """Called just before the ``print`` command is run. It receives the
        list of nodes, and returns the list of nodes that should proceed with
        the command. *id* is a string with the name of the ID to be printed.

        This method can be overridden by derived classes. The default
        implementation does nothing.
        """
        pass

    @doc.api("override")
    def cmd_print_post(self, nodes, id):
        """Called just after the ``print`` command has finished. Arguments are
        as with the ``pre`` method.

        This method can be overridden by derived classes. The default
        implementation does nothing.
        """
        pass

    @doc.api("override")
    def cmd_process_pre(self, trace, options, scripts):
        """Called just before the ``process`` command is run. It receives the
        *trace* to read from as a string, a list of additional Zeek *options*,
        and a list of additional Zeek *scripts*.

        Returns a boolean indicating whether or not the ``process`` command
        should run.

        This method can be overridden by derived classes. The default
        implementation does nothing.
        """
        return True

    @doc.api("override")
    def cmd_process_post(self, trace, options, scripts, success):
        """Called just after the ``process`` command has finished. Arguments
        are as with the ``pre`` method, plus an additional boolean *success*
        indicating whether Zeek terminated normally.

        This method can be overridden by derived classes. The default
        implementation does nothing.
        """
        pass

    # Internal methods.

    def _to_bool(self, val):
        if val.lower() in ("1", "true"):
            return True
        if val.lower() in ("0", "false"):
            return False
        raise ValueError("invalid boolean: '%s'" % val)


    def _registerOptions(self):
        type_converters = {"bool": self._to_bool, "int": int, "string": str}
        pytype = {"bool": bool, "int": int, "string": str}

        for (name, ty, default, descr) in self.options():
            if not name:
                self.error("plugin %s option name must not be empty" % self.name())
                continue

            if "." in name or " " in name:
                self.error('plugin %s option name "%s" must not contain dots or spaces' % (self.name(), name))
                continue

            optname = "%s.%s" % (self.prefix(), name)

            if ty not in pytype:
                self.error('plugin option %s has invalid type "%s"' % (optname, ty))
                continue

            if not isinstance(default, pytype[ty]):
                self.error("plugin option %s default value must be type %s" % (optname, ty))
                continue

            val = config.Config.get_option(optname)
            if val is not None:
                # Convert option values to correct data type for options
                # specified in zeekctl.cfg
                try:
                    newval = type_converters[ty](val)
                except ValueError:
                    self.error('zeekctl option "%s" has invalid value "%s" for type %s' % (optname, val, ty))
                    continue

                config.Config.set_option(optname, newval)
            else:
                # Set default value for options not specified in zeekctl.cfg
                config.Config.init_option(optname, default)




try:
    from . import _broker
except ImportError:
    import _broker

import sys
import datetime
import time
import types
import ipaddress
import collections

try:
    from datetime import timezone
    utc = timezone.utc
except:
    # Only Python 3.2+ has a datetime.timezone.utc we can re-use
    class UTC(datetime.tzinfo):

        def utcoffset(self, dt):
            return datetime.timedelta(0)

        def tzname(self, dt):
            return "UTC"

        def dst(self, dt):
            return datetime.timedelta(0)

        def __eq__(self, other):
            if isinstance(other, UTC):
                return True

            try:
                if self.utcoffset(None) != other.utcoffset(None):
                    return False

                if other.dst(None) is None:
                    return True

                return self.dst(None) == other.dst(None)

            except:
                return False

        def __ne__(self, other):
            return not self.__eq__(other)

    utc = UTC()

Version = _broker.Version
Version.string = lambda: '%u.%u.%u' % (Version.MAJOR, Version.MINOR, Version.PATCH)

now = _broker.now

APIFlags = _broker.APIFlags
EC = _broker.EC
SC = _broker.SC
PeerStatus = _broker.PeerStatus
PeerFlags = _broker.PeerFlags
Frontend = _broker.Frontend
Backend = _broker.Backend
NetworkInfo = _broker.NetworkInfo
EndpointInfo = _broker.EndpointInfo
PeerInfo = _broker.PeerInfo
Topic = _broker.Topic
Status = _broker.Status
Error = _broker.Error
Configuration = _broker.Configuration
BrokerOptions = _broker.BrokerOptions

# Broker's (or better: CAF's) EC code is an integer. Add support
# for comparision against the enum.
_EC_eq = _broker.EC.__eq__
def _our_EC_eq(self, other):
    if isinstance(other, int):
        return other == int(self)
    else:
        return _EC_eq(self, other)

_broker.EC.__eq__ = _our_EC_eq

Address = _broker.Address
Count = _broker.Count
Enum = _broker.Enum
Port = _broker.Port
Set = _broker.Set
Subnet = _broker.Subnet
Table = _broker.Table
Timespan = _broker.Timespan
Timestamp = _broker.Timestamp
Vector = _broker.Vector

def _make_topic(t):
    return (Topic(t) if not isinstance(t, Topic) else t)

def _make_topics(ts):
    if isinstance(ts, Topic):
        ts = [ts]
    elif isinstance(ts, str):
        ts = [Topic(ts)]
    elif isinstance(ts, collections.Iterable):
        ts = [_make_topic(t) for t in ts]
    else:
        ts = [Topic(ts)]

    return _broker.VectorTopic(ts)

# This class does not derive from the internal class because we
# need to pass in existing instances. That means we need to
# wrap all methods, even those that just reuse the internal
# implementation.
class Subscriber:
    def __init__(self, internal_subscriber):
        self._subscriber = internal_subscriber

    def __enter__(self):
        return self

    def __exit__(self, type, value, traceback):
        self._subscriber.reset()
        self._subscriber = None

    def reset(self):
        self._subscriber.reset()
        self._subscriber = None

    def get(self, *args, **kwargs):
        msg = self._subscriber.get(*args, **kwargs)

        if msg is None:
            return None

        if isinstance(msg, _broker.OptionalSubscriberBaseValueType):
            if not msg.is_set():
                return None

            msg = msg.get()

        if isinstance(msg, tuple):
            return (msg[0].string(), Data.to_py(msg[1]))

        if isinstance(msg, _broker.VectorPairTopicData):
            return [(d[0].string(), Data.to_py(d[1])) for d in msg]

        assert False

    def poll(self):
        msgs = self._subscriber.poll()
        return [(d[0].string(), Data.to_py(d[1])) for d in msgs]

    def available(self):
        return self._subscriber.available()

    def fd(self):
        return self._subscriber.fd()

    def add_topic(self, topic, block=False):
        return self._subscriber.add_topic(_make_topic(topic), block)

    def remove_topic(self, topic, block=False):
        return self._subscriber.remove_topic(_make_topic(topic), block)

class SafeSubscriber(Subscriber):
    """Subscriber subclass that makes returnes messages safe to process.

    "Safe" here means safe to Python's type system, particularly regarding
    hashable types. Broker's data model permits nested complex types, such as
    sets of tables, but those don't directly work in Python (for example,
    constructing a set of dicts will complain that dicts aren't hashable). To
    work around this SafeSubscriber relies on ImmutableData rather than Data
    (used by regular Subscribers)."""

    def get(self, *args, **kwargs):
        msg = self._subscriber.get(*args, **kwargs)

        if msg is None:
            return None

        if isinstance(msg, _broker.OptionalSubscriberBaseValueType):
            if not msg.is_set():
                return None

            msg = msg.get()

        if isinstance(msg, tuple):
            return (msg[0].string(), ImmutableData.to_py(msg[1]))

        if isinstance(msg, _broker.VectorPairTopicData):
            return [(d[0].string(), ImmutableData.to_py(d[1])) for d in msg]

        assert False

class StatusSubscriber():
    def __init__(self, internal_subscriber):
        self._subscriber = internal_subscriber

    def __enter__(self):
        return self

    def __exit__(self, type, value, traceback):
        self._subscriber.reset()
        self._subscriber = None

    def reset(self):
        self._subscriber.reset()
        self._subscriber = None

    def get(self, *args, **kwargs):
        x = self._subscriber.get(*args, **kwargs)
        return self._to_result(x)

    def poll(self):
        xs = self._subscriber.poll()
        return [self._to_result(x) for x in xs]

    def available(self):
        return self._subscriber.available()

    def fd(self):
        return self._subscriber.fd()

    def _to_result(self, x):
        if isinstance(x, _broker.VectorStatusSubscriberValueType):
            return [self._to_error_or_status(xi) for xi in x]

        return self._to_error_or_status(x)

    def _to_error_or_status(self, x):
        if x.is_error():
            return x.get_error()

        if x.is_status():
            return x.get_status()

        assert False

class Publisher:
    # This class does not derive from the internal class because we
    # need to pass in existing instances. That means we need to
    # wrap all methods, even those that just reuse the internal
    # implementation.
    def __init__(self, internal_publisher):
        self._publisher = internal_publisher

    def __enter__(self):
        return self

    def __exit__(self, type, value, traceback):
        self._publisher.reset()
        self._publisher = None

    def reset(self):
        self._publisher.reset()
        self._publisher = None

    def demand(self):
        return self._publisher.demand()

    def buffered(self):
        return self._publisher.buffered()

    def capacity(self):
        return self._publisher.capacity()

    def free_capacity(self):
        return self._publisher.free_capacity()

    def send_rate(self):
        return self._publisher.send_rate()

    def fd(self):
        return self._publisher.fd()

    def publish(self, data):
        data = Data.from_py(data)
        return self._publisher.publish(data)

    def publish_batch(self, *batch):
        batch = [Data.from_py(d) for d in batch]
        return self._publisher.publish_batch(_broker.Vector(batch))

class Store:
    # This class does not derive from the internal class because we
    # need to pass in existing instances. That means we need to
    # wrap all methods, even those that just reuse the internal
    # implementation.
    def __init__(self, internal_store):
        self._store = internal_store
        # Points to the "owning" Endpoint to make sure Python cleans this object up
        # before destroying the endpoint.
        self._parent = None

    def __enter__(self):
        return self

    def __exit__(self, type, value, traceback):
        self._store.reset()
        self._parent = None
        self._store = None

    def __enter__(self):
        return self

    def __exit__(self, type, value, traceback):
        self._store.reset()
        self._parent = None
        self._store = None

    def name(self):
        return self._store.name()

    def exists(self, key):
        key = Data.from_py(key)
        value = self._store.exists(key)
        return Data.to_py(value.get())

    def get(self, key):
        key = Data.from_py(key)
        value = self._store.get(key)
        return Data.to_py(value.get()) if value.is_valid() else None

    def get_index_from_value(self, key, index):
        key = Data.from_py(key)
        index = Data.from_py(index)
        value = self._store.get_index_from_value(key, index)
        return Data.to_py(value.get()) if value.is_valid() else None

    def keys(self):
        keys = self._store.keys()

        if keys is None:
            return None

        return Data.to_py(keys.get()) if keys.is_valid() else None

    def put(self, key, value, expiry=None):
        key = Data.from_py(key)
        value = Data.from_py(value)
        expiry = self._to_expiry(expiry)
        return self._store.put(key, value, expiry)

    def put_unique(self, key, value, expiry=None):
        key = Data.from_py(key)
        value = Data.from_py(value)
        expiry = self._to_expiry(expiry)
        rval = self._store.put_unique(key, value, expiry)
        return Data.to_py(rval.get()) if rval.is_valid() else None

    def erase(self, data):
        data = Data.from_py(data)
        return self._store.erase(data)

    def clear(self):
        return self._store.clear()

    def increment(self, key, amount, expiry=None):
        key = Data.from_py(key)
        amount = Data.from_py(amount)
        expiry = self._to_expiry(expiry)
        return self._store.increment(key, amount, expiry)

    def decrement(self, key, amount, expiry=None):
        key = Data.from_py(key)
        amount = Data.from_py(amount)
        expiry = self._to_expiry(expiry)
        return self._store.decrement(key, amount, expiry)

    def append(self, key, s, expiry=None):
        key = Data.from_py(key)
        s = Data.from_py(s)
        expiry = self._to_expiry(expiry)
        return  self._store.append(key, s, expiry)

    def insert_into(self, key, index, value=None, expiry=None):
        key = Data.from_py(key)
        index = Data.from_py(index)
        expiry = self._to_expiry(expiry)

        if value:
            value = Data.from_py(value)
            return self._store.insert_into(key, index, value, expiry)
        else:
            return self._store.insert_into(key, index, expiry)

    def remove_from(self, key, index, expiry=None):
        key = Data.from_py(key)
        index = Data.from_py(index)
        expiry = self._to_expiry(expiry)
        return  self._store.remove_from(key, index, expiry)

    def push(self, key, value, expiry=None):
        key = Data.from_py(key)
        value = Data.from_py(value)
        expiry = self._to_expiry(expiry)
        return self._store.push(key, value, expiry)

    def pop(self, key, expiry=None):
        key = Data.from_py(key)
        expiry = self._to_expiry(expiry)
        return self._store.pop(key, expiry)

    def _to_expiry(self, e):
        return (_broker.OptionalTimespan(_broker.Timespan(float(e))) if e is not None else _broker.OptionalTimespan())

    def await_idle(self, timeout=None):
        if timeout:
            return self._store.await_idle(_broker.Timespan(float(timeout)))
        else:
            return self._store.await_idle()

    # Points to the "owning" Endpoint to make sure Python cleans this object up
    # before destroying the endpoint.
    _parent = None

class Endpoint(_broker.Endpoint):
    def make_subscriber(self, topics, qsize = 20, subscriber_class=Subscriber):
        topics = _make_topics(topics)
        s = _broker.Endpoint.make_subscriber(self, topics, qsize)
        return subscriber_class(s)

    def make_safe_subscriber(self, topics, qsize = 20):
        """A variant of make_subscriber that returns a SafeSubscriber instance. In
        contrast to the Subscriber class, messages retrieved from
        SafeSubscribers use immutable, hashable values to ensure Python can
        represent them. When in doubt, use make_safe_subscriber()."""
        return self.make_subscriber(topics=topics, qsize=qsize, subscriber_class=SafeSubscriber)

    def make_status_subscriber(self, receive_statuses=False):
        s = _broker.Endpoint.make_status_subscriber(self, receive_statuses)
        return StatusSubscriber(s)

    def make_publisher(self, topic):
        topic = _make_topic(topic)
        p = _broker.Endpoint.make_publisher(self, topic)
        return Publisher(p)

    def forward(self, topics):
        topics = _make_topics(topics)
        _broker.Endpoint.forward(self, topics)

    def publish(self, topic, data):
        topic = _make_topic(topic)
        data =  Data.from_py(data)
        return _broker.Endpoint.publish(self, topic, data)

    def publish_batch(self, *batch):
        batch = [(_make_topic(t), Data.from_py(d)) for (t, d) in batch]
        return _broker.Endpoint.publish_batch(self, _broker.VectorPairTopicData(batch))

    def attach_master(self, name, type=None, opts={}):
        bopts = _broker.MapBackendOptions() # Generator expression doesn't work here.
        for (k, v) in opts.items():
            bopts[k] = Data.from_py(v)
        s = _broker.Endpoint.attach_master(self, name, type, bopts)
        if not s.is_valid():
            return None
        result = Store(s.get())
        # This extra reference establishes a parent-child relation between the
        # two Python objects, making sure that the garbage collector destroys
        # the store *before* cleaning up the endpoint
        result._parent = self
        return result

    def attach_clone(self, name):
        s = _broker.Endpoint.attach_clone(self, name)
        if not s.is_valid():
            return None
        result = Store(s.get())
        # Same as above: make sure Python cleans up the store first.
        result._parent = self
        return result

    def await_peer(self, node, timeout=None):
        if timeout:
            return  _broker.Endpoint.await_peer(self, node, _broker.Timespan(float(timeout)))
        else:
            return  _broker.Endpoint.await_peer(self, node)

    def __enter__(self):
        return self

    def __exit__(self, type, value, traceback):
        self.shutdown()

class Message:
    def to_broker(self):
        assert False and "method not overridden"

from . import zeek

class Data(_broker.Data):
    def __init__(self, x = None):
        if x is None:
            _broker.Data.__init__(self)

        elif isinstance(x, zeek.Event):
            _broker.Data.__init__(self, x.as_data())

        elif isinstance(x, _broker.Data):
            _broker.Data.__init__(self, x)

        elif isinstance(x, (bool, int, float, str, bytes,
                            Address, Count, Enum, Port, Set, Subnet, Table, Timespan, Timestamp, Vector)):
            _broker.Data.__init__(self, x)

        elif isinstance(x, datetime.timedelta):
            us = x.microseconds + (x.seconds + x.days * 24 * 3600) * 10**6
            ns = us * 10**3
            _broker.Data.__init__(self, _broker.Timespan(ns))

        elif isinstance(x, datetime.datetime):
            secs = x.timestamp()
            _broker.Data.__init__(self, _broker.Timestamp(secs))

        elif isinstance(x, ipaddress.IPv4Address):
            _broker.Data.__init__(self, _broker.Address(x.packed, 4))

        elif isinstance(x, ipaddress.IPv6Address):
            _broker.Data.__init__(self, _broker.Address(x.packed, 6))

        elif isinstance(x, ipaddress.IPv4Network):
            address = _broker.Address(x.network_address.packed, 4)
            length = x.prefixlen
            _broker.Data.__init__(self, _broker.Subnet(address, length))

        elif isinstance(x, ipaddress.IPv6Network):
            address = _broker.Address(x.network_address.packed, 6)
            length = x.prefixlen
            _broker.Data.__init__(self, _broker.Subnet(address, length))

        elif isinstance(x, list) or isinstance(x, tuple):
            v = _broker.Vector([Data(i) for i in x])
            _broker.Data.__init__(self, v)

        elif isinstance(x, set) or isinstance(x, frozenset):
            s = _broker.Set(([Data(i) for i in x]))
            _broker.Data.__init__(self, s)

        elif isinstance(x, dict) or isinstance(x, types.MappingProxyType):
            t = _broker.Table()
            for (k, v) in x.items():
                t[Data(k)] = Data(v)

            _broker.Data.__init__(self, t)

        else:
            raise TypeError("unsupported data type: " + str(type(x)))

    @staticmethod
    def from_py(x):
        return Data(x)

    @staticmethod
    def to_py(d):
        def to_ipaddress(a):
            if a.is_v4():
                return ipaddress.IPv4Address(a.bytes()[-4:])
            else:
                return ipaddress.IPv6Address(a.bytes())

        def to_subnet(s):
            # Python < 3.5 does not have a nicer way of setting the prefixlen
            # when creating from packed data.
            if s.network().is_v4():
                return ipaddress.IPv4Network(to_ipaddress(s.network())).supernet(new_prefix=s.length())
            else:
                return ipaddress.IPv6Network(to_ipaddress(s.network())).supernet(new_prefix=s.length())

        def to_set(s):
            return set([Data.to_py(i) for i in s])

        def to_table(t):
            return {Data.to_py(k): Data.to_py(v) for (k, v) in t.items()}

        def to_vector(v):
            return tuple(Data.to_py(i) for i in v)

        def _try_bytes_decode(b):
            try:
                return b.decode('utf-8')
            except:
                return b

        converters = {
            Data.Type.Nil: lambda: None,
            Data.Type.Address: lambda: to_ipaddress(d.as_address()),
            Data.Type.Boolean: lambda: d.as_boolean(),
            Data.Type.Count: lambda: Count(d.as_count()),
            Data.Type.EnumValue: lambda: d.as_enum_value(),
            Data.Type.Integer: lambda: d.as_integer(),
            Data.Type.Port: lambda: d.as_port(),
            Data.Type.Real: lambda: d.as_real(),
            Data.Type.Set: lambda: to_set(d.as_set()),
            Data.Type.String: lambda: _try_bytes_decode(d.as_string()),
            Data.Type.Subnet: lambda: to_subnet(d.as_subnet()),
            Data.Type.Table: lambda: to_table(d.as_table()),
            Data.Type.Timespan: lambda: datetime.timedelta(seconds=d.as_timespan()),
            Data.Type.Timestamp: lambda: datetime.datetime.fromtimestamp(d.as_timestamp(), utc),
            Data.Type.Vector: lambda: to_vector(d.as_vector())
        }

        try:
            return converters[d.get_type()]()
        except KeyError:
            raise TypeError("unsupported data type: " + str(d.get_type()))

class ImmutableData(Data):
    """A Data specialization that uses immutable complex types for returned Python
    objects. For sets, the return type is frozenset, for tables it's a hashable,
    read-only derivative of dict, and for vectors it's Python tuples.
    """
    class HashableReadOnlyDict(dict):
        def __hash__(self):
            return hash(frozenset(self.items()))

        def __readonly__(self, *args, **kwargs):
            raise TypeError('cannot modify this dict')

        # https://stackoverflow.com/a/31049908
        __setitem__ = __readonly__
        __delitem__ = __readonly__
        pop = __readonly__
        popitem = __readonly__
        clear = __readonly__
        update = __readonly__
        setdefault = __readonly__
        del __readonly__

    @staticmethod
    def to_py(d):
        def to_set(s):
            return frozenset([ImmutableData.to_py(i) for i in s])

        def to_table(t):
            tmp = {ImmutableData.to_py(k): ImmutableData.to_py(v) for (k, v) in t.items()}
            # It's tempting here to use types.MappingProxyType here, but it is
            # not hashable, so doesn't solve our main problem, and cannot be
            # derived from.
            return ImmutableData.HashableReadOnlyDict(tmp.items())

        def to_vector(v):
            return tuple(ImmutableData.to_py(i) for i in v)

        converters = {
            Data.Type.Set: lambda: to_set(d.as_set()),
            Data.Type.Table: lambda: to_table(d.as_table()),
            Data.Type.Vector: lambda: to_vector(d.as_vector())
        }

        try:
            return converters[d.get_type()]()
        except KeyError:
            # Fall back on the Data class for types we handle identically.
            return Data.to_py(d)

####### TODO: Updated to new Broker API until here.

# # TODO: complete interface
# class Store:
#   def __init__(self, handle):
#     self.store = handle
#
#   def name(self):
#     return self.store.name()
#
# class Mailbox:
#   def __init__(self, handle):
#     self.mailbox = handle
#
#   def descriptor(self):
#     return self.mailbox.descriptor()
#
#   def empty(self):
#     return self.mailbox.empty()
#
#   def count(self, n = -1):
#     return self.mailbox.count(n)
#
#
# class Message:
#   def __init__(self, handle):
#     self.message = handle
#
#   def topic(self):
#     return self.message.topic().string()
#
#   def data(self):
#     return self.message.data() # TODO: unwrap properly
#
#   def __str__(self):
#     return "%s -> %s" % (self.topic(), str(self.data()))
#
#
# class BlockingEndpoint(Endpoint):
#   def __init__(self, handle):
#     super(BlockingEndpoint, self).__init__(handle)
#
#   def subscribe(self, topic):
#     self.endpoint.subscribe(topic)
#
#   def unsubscribe(self, topic):
#     self.endpoint.unsubscribe(topic)
#
#   def receive(self, x):
#     if x == Status:
#       return self.endpoint.receive()
#     elif x == Message:
#       return Message(self.endpoint.receive())
#     else:
#       raise BrokerError("invalid receive type")
#
#   #def receive(self):
#   #  if fun1 is None:
#   #    return Message(self.endpoint.receive())
#   #  if fun2 is None:
#   #    if utils.arity(fun1) == 1:
#   #      return self.endpoint.receive_status(fun1)
#   #    if utils.arity(fun1) == 2:
#   #      return self.endpoint.receive_msg(fun1)
#   #    raise BrokerError("invalid receive callback arity; must be 1 or 2")
#   #  return self.endpoint.receive_msg_or_status(fun1, fun2)
#
#   def mailbox(self):
#     return Mailbox(self.endpoint.mailbox())
#
#
# class NonblockingEndpoint(Endpoint):
#   def __init__(self, handle):
#     super(NonblockingEndpoint, self).__init__(handle)
#
#   def subscribe(self, topic, fun):
#     self.endpoint.subscribe_msg(topic, fun)
#
#   def on_status(fun):
#     self.endpoint.subscribe_status(fun)
#
#   def unsubscribe(self, topic):
#     self.endpoint.unsubscribe(topic)
#
#
# class Context:
#   def __init__(self):
#     self.context = _broker.Context()
#
#   def spawn(self, api):
#     if api == Blocking:
#       return BlockingEndpoint(self.context.spawn_blocking())
#     elif api == Nonblocking:
#       return NonblockingEndpoint(self.context.spawn_nonblocking())
#     else:
#       raise BrokerError("invalid API flag: " + str(api))
#



"""This module provides Broker-based communication with a Zeek cluster controller."""
import ssl
import sys
import time

# This requires the websockets-client package, confusingly imported under the
# name "websocket", and called "python3-websocket" in the Debian/Ubuntu
# world. Not to be confused with "websockets", the alternative, async-oriented
# Python package! Please see: https://websocket-client.readthedocs.io
try:
    import websocket
except ImportError:
    print(
        "The zeek-client package requires websocket-client.\n"
        "If you use 'pip', you can install it as follows:\n"
        "\n"
        "    pip3 install websocket-client\n"
        "\n",
        file=sys.stderr,
    )
    sys.exit(1)

from .brokertypes import (
    ZeekEvent,
    HandshakeMessage,
    HandshakeAckMessage,
    DataMessage,
)

from .config import CONFIG
from .consts import CONTROLLER_TOPIC
from .events import Registry
from .logs import LOG
from .ssl import get_websocket_sslopt
from .utils import make_uuid


class Error(Exception):
    """Catch-all for exceptions arising from use of Controller objects."""


class ConfigError(Error):
    """A problem occurred while configuring the WebSocket object."""


class UsageError(Error):
    """Invalid sequence of operations on a Controller object."""


class Controller:
    """A class managing a connection to the Zeek cluster controller."""

    def __init__(
        self,
        controller_host=None,
        controller_port=None,
        controller_topic=CONTROLLER_TOPIC,
    ):
        """Controller connection constructor.

        This may raise ConfigError in case of trouble with the
        connection settings.
        """
        self.controller_host = controller_host or CONFIG.get("controller", "host")
        self.controller_port = controller_port or CONFIG.getint("controller", "port")
        self.controller_topic = controller_topic
        self.controller_broker_id = None  # Defined in Handshake ACK message

        try:
            if self.controller_port < 1 or self.controller_port > 65535:
                raise ValueError(
                    f"controller port number {self.controller_port} outside valid range"
                )

            disable_ssl = CONFIG.getboolean("ssl", "disable")

            proto = "ws" if disable_ssl else "wss"
            remote = f"{self.controller_host}:{self.controller_port}"
            self.wsock_url = f"{proto}://{remote}/v1/messages/json"

            sslopt = None if disable_ssl else get_websocket_sslopt()
            self.wsock = websocket.WebSocket(sslopt=sslopt)
        except (ValueError, OSError, ssl.SSLError) as err:
            raise ConfigError(
                f"cannot configure connection to "
                f"{self.controller_host}:{self.controller_port}: {err}"
            ) from err

    def connect(self):
        """Connect to the configured controller.

        This takes the controller coordonates from the zeek-client configuration
        (or the arguments passed to the constructor, if any) and establishes a
        fully peered connection. "Fully peered" here means that the object first
        establishes the websocket connection, potentially wrapped in TLS as per
        the TLS-specific configuration settings, and then conducts the
        Broker-level handshake. The latter establishes the Controller's Broker
        ID and our topic subscriptions.

        Returns True if peering completes successfully, False otherwise, with
        according messages written to the log.
        """
        LOG.info(
            "connecting to controller %s:%s", self.controller_host, self.controller_port
        )

        attempts = CONFIG.getint("client", "peering_attempts")
        retry_delay = CONFIG.getfloat("client", "peering_retry_delay_secs")

        handshake = HandshakeMessage([self.controller_topic])

        # We accommodate problems during connect() and the Broker handshake,
        # attempting these a total of client.peering_attempts times.  That is,
        # if we use 10 attempts and connect() takes 3 attempts, 7 attempts
        # remain for the handshake. Since the kinds of problems that may arise
        # in either stage in the (web)socket operations overlap substantially,
        # we use a single function that checks them all:
        def wsock_operation(op, stage):
            nonlocal attempts

            while attempts > 0:
                try:
                    attempts -= 1
                    return op()
                except websocket.WebSocketTimeoutException:
                    time.sleep(retry_delay)
                    continue
                except websocket.WebSocketException as err:
                    LOG.error(
                        "websocket error in %s with controller %s:%s: %s",
                        stage,
                        self.controller_host,
                        self.controller_port,
                        err,
                    )
                    return False
                except ConnectionRefusedError:
                    # We don't consider these fatal since they can happen
                    # naturally during tests and other automated setups where
                    # it's beneficial to keep trying.  Also, this is a subclass
                    # of OSError, so needs to come before it:
                    LOG.debug(
                        "connection refused for controller %s:%s",
                        self.controller_host,
                        self.controller_port,
                    )
                    time.sleep(retry_delay)
                    continue
                except ssl.SSLError as err:
                    # Same here, likewise a subclass of OSError:
                    LOG.error(
                        "socket TLS error in %s with controller %s:%s: %s",
                        stage,
                        self.controller_host,
                        self.controller_port,
                        err,
                    )
                    return False
                except OSError as err:
                    # From socket.py docs: "Errors related to socket or address
                    # semantics raise OSError or one of its subclasses".
                    LOG.error(
                        "socket error in %s with controller %s:%s: %s",
                        stage,
                        self.controller_host,
                        self.controller_port,
                        err,
                    )
                    return False
                except Exception as err:
                    LOG.exception(
                        "unexpected error in %s with controller %s:%s: %s",
                        stage,
                        self.controller_host,
                        self.controller_port,
                        err,
                    )
                    return False

            if attempts == 0:
                LOG.error(
                    "websocket connection to %s:%s timed out in %s",
                    self.controller_host,
                    self.controller_port,
                    stage,
                )
            return False

        def connect_op():
            self.wsock.connect(self.wsock_url, timeout=retry_delay)
            self.wsock.send(handshake.serialize())
            return True

        def handshake_op():
            rawdata = self.wsock.recv()
            try:
                msg = HandshakeAckMessage.unserialize(rawdata)
            except TypeError as err:
                LOG.error(
                    "protocol data error with controller %s:%s: %s, raw data: %s",
                    self.controller_host,
                    self.controller_port,
                    err,
                    rawdata,
                )
                return False

            self.controller_broker_id = msg.endpoint
            LOG.info(
                "peered with controller %s:%s",
                self.controller_host,
                self.controller_port,
            )
            return True

        if not wsock_operation(connect_op, "connect()"):
            return False
        if not wsock_operation(handshake_op, "handshake"):
            return False

        return True

    def publish(self, event):
        """Publishes the given event to the controller topic.

        Raises UsageError when invoked without an earlier connect().

        Args:
            event (zeekclient.event.Event): the event to publish.
        """
        if self.controller_broker_id is None:
            raise UsageError("cannot publish without established peering")

        msg = DataMessage(self.controller_topic, event.to_brokertype())
        self.wsock.send(msg.serialize())

    def receive(self, timeout_secs=None, filter_pred=None):
        """Receive an event from the controller's event subscriber.

        Raises UsageError when invoked without an earlier connect().

        Args:
            timeout_secs (int): number of seconds before we time out.
                Has sematics of the poll.poll() timeout argument, i.e.
                None and negative values mean no timeout. The default
                is client.request_timeout_secs.

            filter_pred: a predicate function for filtering out unacceptable
                events. The function takes a received event as only input,
                returning True if the event is acceptable for returning to the
                `receive()` caller, and False otherwise. When not provided,
                any received event is acceptable. When the predicate returns
                false, the wait for a suitable event continues, subject to the
                same overall timeout.

        Returns:
            A tuple of (1) an instance of one of the Event classes defined for
            the client, or None if timeout_secs passed before anything arrived,
            and (2) a string indicating any occurring errors. The string is
            empty when no error occurs.
        """
        if self.controller_broker_id is None:
            raise UsageError("cannot receive without established peering")

        timeout = timeout_secs or CONFIG.getint("client", "request_timeout_secs")
        old_timeout = self.wsock.gettimeout()

        try:
            self.wsock.settimeout(timeout)

            remote = f"{self.controller_host}:{self.controller_port}"

            while True:
                # Reading the event proceeds in three steps:
                # (1) read data from the websocket
                # (2) ensure it's a data message
                # (3) try to extract data message payload as event
                try:
                    msg = DataMessage.unserialize(self.wsock.recv())
                except TypeError as err:
                    return (
                        None,
                        f"protocol data error with controller {remote}: {err}",
                    )
                except websocket.WebSocketTimeoutException:
                    return (
                        None,
                        f"websocket connection to {remote} timed out",
                    )
                except Exception as err:
                    LOG.exception("unexpected error")
                    return (
                        None,
                        f"unexpected error with controller {remote}: {err}",
                    )
                try:
                    # Events are a specially laid-out vector of vectors:
                    # https://docs.zeek.org/projects/broker/en/current/web-socket.html#encoding-of-zeek-events
                    evt = ZeekEvent.from_vector(msg.data)

                    # Turn Broker-level event data into a zeekclient.event.Event:
                    res = Registry.make_event(evt.name, *evt.args)
                    if res is not None and (filter_pred is None or filter_pred(res)):
                        return res, ""
                except TypeError as err:
                    return None, (
                        f"protocol data error with controller {remote}: "
                        f"invalid event data, {repr(msg.data)}"
                    )

                # This wasn't the event type we wanted, try again.
        finally:
            self.wsock.settimeout(old_timeout)

    def transact(self, request_type, response_type, *request_args, reqid=None):
        """Pairs publishing a request event with receiving its response event.

        This is a wrapper around :meth:`.Controller.publish()` with subsequent
        :meth:`.Controller.receive()`, with automatic provision of a request ID
        in the request event, and validation of a matching request ID in the
        response. Mismatching response events are ignored, and lack of a
        suitable event in the timeout period leads to an empty result with
        according error message, just like :meth:`.Controller.receive()`.

        The function works only with request and response event types that take
        a "reqid" string as first argument. The function verifies this lightly,
        just by looking at the name of the first argument. See
        `zeekclient.events` for suitable event types.

        Raises UsageError when invoked without an earlier connect().

        Args:
            request_type (zeekclient.event.Event class): the request event type.

            response_type (zeekclient.event.Event class): the response event type.

            request_args: any event arguments in addition to the initial "reqid" string.

            reqid (str): the request ID to use in the request event, and expect
                in the response event. When omitted, the function produces its
                own ID.

        Returns:
            The same as Controller.receive(): tuple of an event instance
            and a string indicating any error.
        """
        # Verify that the first arguments of the event types are actually a
        # request ID -- we just look at the name:
        if request_type.ARG_NAMES[0] != "reqid":
            return (
                None,
                f"type error: event type {request_type.__name__} does not have request ID",
            )
        if response_type.ARG_NAMES[0] != "reqid":
            return (
                None,
                f"type error: event type {response_type.__name__} does not have request ID",
            )

        if reqid is None:
            reqid = make_uuid()

        evt = request_type(reqid, *request_args)
        self.publish(evt)

        def is_response(evt):
            try:
                return isinstance(evt, response_type) and evt.reqid.to_py() == reqid
            except AttributeError:
                return False

        return self.receive(filter_pred=is_response)



import os
import errno

from ZeekControl import config


# 'src' is the file to which the link will point, and 'dst' is the link to make
def force_symlink(src, dst):
    try:
        os.symlink(src, dst)
    except OSError as e:
        if e.errno == errno.EEXIST:
            os.remove(dst)
            os.symlink(src, dst)
        else:
            raise

# Returns an IP address string suitable for embedding in a Zeek script,
# for IPv6 colon-hexadecimal address strings, that means surrounding it
# with square brackets.
def format_zeek_addr(addr):
    if ":" not in addr:
        return addr
    else:
        return "[%s]" % addr

# Returns an IP prefix string suitable for embedding in a Zeek script,
# for IPv6 colon-hexadecimal prefix strings, that means surrounding the
# IP address part with square brackets.
def format_zeek_prefix(prefix):
    if ":" not in prefix:
        return prefix
    else:
        parts = prefix.split("/")
        return "[%s]/%s" % (parts[0], parts[1])

# Returns an IP address string suitable for use with rsync, which requires
# encasing IPv6 addresses in square brackets, and some shells may require
# quoting the brackets.
def format_rsync_addr(addr):
    if ":" not in addr:
        return addr
    else:
        return "'[%s]'" % addr

# Convert a number into a string with a unit (e.g., 1024 into "1K").
def number_unit_str(num):
    units = (("G", 1024*1024*1024), ("M", 1024*1024), ("K", 1024))
    for (unit, factor) in units:
        if num >= factor:
            return "%3.0f%s" % (num / factor, unit)
    return " %3.0f" % (num)




from __future__ import print_function
import sys

from ZeekControl import options
from ZeekControl import doc
from ZeekControl import plugin
from ZeekControl import node

# Prints the command docstrings in a form suitable for direct inclusion
# into the documentation.
def print_commands(cls):
    cmds = []

    for i in cls.__dict__:
        docstr = cls.__dict__[i].__doc__
        if i.startswith("do_") and docstr:
            cmds += [(i[3:], docstr)]

    cmds.sort()

    for (cmd, docstr) in cmds:
        if docstr.startswith("- "):
            # First line are arguments.
            docstr = docstr.splitlines()
            args = docstr[0][2:]
            docstr = "\n".join(docstr[1:])
        else:
            args = ""

        if args:
            args = " *%s*" % args

        output = ""
        for line in docstr.splitlines():
            output += "    %s\n" % line.strip()

        output = output.strip()

        print()
        print(".. _%s:\n\n*%s*%s\n    %s" % (cmd, cmd, args, output))
        print()


# Print options documentation.
def print_options():
    print("User Options")
    print("~~~~~~~~~~~~")

    out, err = options.print_options(options.Option.USER)
    print(out, end="")
    if err:
        print(err, file=sys.stderr)

    print()
    print("Internal Options")
    print("~~~~~~~~~~~~~~~~")
    print()

    out, err = options.print_options(options.Option.AUTOMATIC)
    print(out, end="")
    if err:
        print(err, file=sys.stderr)


# Print plugin and node documentation.
def print_plugin():
    print(doc.print_class(plugin.Plugin, tag="no-methods"), end="")
    print(doc.print_class(plugin.Plugin, header=False), end="")
    print(doc.print_class(plugin.Plugin, "override", header=False), end="")

    print(doc.print_class(node.Node), end="")

def print_zeekctl_docs(mainpath, zeekctlclass):
    print(".. Autogenerated. Do not edit.\n")
    with open(mainpath, "r") as f:
        for line in f:
            fields = line.strip().split(None, 2)
            if len(fields) == 3 and fields[0] == ".." and fields[1] == "include::":
                if fields[2] == "commands.rst":
                    print_commands(zeekctlclass)
                elif fields[2] == "options.rst":
                    print_options()
                elif fields[2] == "plugins.rst":
                    print_plugin()
            else:
                print(line, end="")




# ZeekControl exceptions.
#
# These are the exceptions that are expected to be raised under predictable
# circumstances (such as bad user input, or an invalid configuration).  When
# these are raised, a text message explaining the problem should be provided.
# Therefore, these should be caught to avoid seeing a stack trace (which
# provides no new information in these cases).  However, standard Python
# exceptions are not expected to occur, so if one is raised a stack trace
# can provide valuable information on the source of the problem.

class ZeekControlError(Exception):
    """This is the base class for ZeekControl exceptions."""

class LockError(ZeekControlError):
    """Indicates that ZeekControl was unable to obtain a lock."""

class RuntimeEnvironmentError(ZeekControlError):
    """Indicates an error in the runtime environment (e.g. running as wrong
    user, or some files/directories have wrong permissions or location)."""

class InvalidNodeError(ZeekControlError):
    """Indicates an attempt to lookup an invalid node name."""

class ConfigurationError(ZeekControlError):
    """Indicates a problem with the ZeekControl configuration."""

class CommandSyntaxError(ZeekControlError):
    """Indicates a syntax error in a ZeekControl command."""



#
# A simple test plugin exercising most of a plugins capabilities. It just does
# some debugging output, nothing else.
#
# The plugin is off by default. To enable it, add "test.enabled=1" to zeekctl.cfg.

import ZeekControl.plugin
import ZeekControl.cmdresult

class TestPlugin(ZeekControl.plugin.Plugin):
    def __init__(self):
        super(TestPlugin, self).__init__(apiversion=1)

    def name(self):
        return "TestPlugin"

    def prefix(self):
        return "test"

    def pluginVersion(self):
        return 1

    def init(self):
        if not self.getOption("enabled"):
            return False

        foo = self.getOption("foo")

        self.message("TestPlugin: Test initialized")
        self.message("TestPlugin: The value of foo is: %s" % foo)
        self.message("TestPlugin: The current value of bar is: %s" % self.getState("bar"))

        for n in self.nodes():
            try:
                self.message("TestPlugin: mykey is: %s" % n.test_mykey)
            except AttributeError:
                self.message("TestPlugin: mykey is not set")

        for h in self.hosts():
            self.message("TestPlugin: host %s" % h.host)

        return True

    def options(self):
        return [("foo", "int", 1, "Just a test option."),
                ("enabled", "bool", False, "Set to enable plugin")]

    def commands(self):
        return [("bar", "", "A test command from the Test plugin.")]

    def nodeKeys(self):
        return ["mykey"]

    def zeekctl_config(self):
        script = "# This is a test."
        return script

    def done(self):
        self.message("TestPlugin: done")

    def _nodes(self, nodes):

        if not nodes:
            return "<empty>"

        if isinstance(nodes[0], tuple):
            nodes = [n[0] for n in nodes]

        return ",".join([str(n) for n in nodes])

    def _results(self, results):

        if not results:
            return "<empty>"

        return ",".join(["%s/%s" % (str(n[0]), n[1]) for n in results])

    def zeekProcessDied(self, node):
        self.message("TestPlugin: Zeek process died for %s" % node)

    def hostStatusChanged(self, host, status):
        self.message("TestPlugin: host status changed: %s -> %s" % (host, status))

    def cmd_custom(self, cmd, args, cmdout):
        results = ZeekControl.cmdresult.CmdResult()

        bar = self.getState("bar")
        if not bar:
            bar = 1

        self.setState("bar", bar + 1)
        self.message("TestPlugin: My command: %s" % args)

        return results

    def cmd_check_pre(self, nodes):
        self.message("TestPlugin: Test pre 'check':  %s" % self._nodes(nodes))

    def cmd_check_post(self, results):
        self.message("TestPlugin: Test post 'check': %s" % self._results(results))

    def cmd_nodes_pre(self):
        self.message("TestPlugin: Test pre 'nodes'")
        return True

    def cmd_nodes_post(self):
        self.message("TestPlugin: Test post 'nodes'")

    def cmd_config_pre(self):
        self.message("TestPlugin: Test pre 'config'")
        return True

    def cmd_config_post(self):
        self.message("TestPlugin: Test post 'config'")

    def cmd_deploy_pre(self):
        self.message("TestPlugin: Test pre 'deploy'")
        return True

    def cmd_deploy_post(self):
        self.message("TestPlugin: Test post 'deploy'")

    def cmd_exec_pre(self, cmdline):
        self.message("TestPlugin: Test pre 'exec':  %s" % cmdline)
        return True

    def cmd_exec_post(self, cmdline):
        self.message("TestPlugin: Test post 'exec': %s" % cmdline)

    def cmd_install_pre(self):
        self.message("TestPlugin: Test pre 'install'")
        return True

    def cmd_install_post(self):
        self.message("TestPlugin: Test post 'install'")

    def cmd_cron_pre(self, arg, watch):
        self.message("TestPlugin: Test pre 'cron':  %s/%s" % (arg, watch))
        return True

    def cmd_cron_post(self, arg, watch):
        self.message("TestPlugin: Test post 'cron': %s/%s" % (arg, watch))

    def cmd_restart_pre(self, nodes, clean):
        self.message("TestPlugin: Test pre 'restart':  %s (%s)" % (self._nodes(nodes), clean))

    def cmd_restart_post(self, nodes):
        self.message("TestPlugin: Test post 'restart': %s" % self._nodes(nodes))

    def cmd_start_pre(self, nodes):
        self.message("TestPlugin: Test pre 'start':  %s" % self._nodes(nodes))

    def cmd_start_post(self, results):
        self.message("TestPlugin: Test post 'start': %s" % self._results(results))

    def cmd_stop_pre(self, nodes):
        self.message("TestPlugin: Test pre 'stop':  %s" % self._nodes(nodes))

    def cmd_stop_post(self, results):
        self.message("TestPlugin: Test post 'stop': %s" % self._results(results))

    def cmd_status_pre(self, nodes):
        self.message("TestPlugin: Test pre 'status':  %s" % self._nodes(nodes))

    def cmd_status_post(self, nodes):
        self.message("TestPlugin: Test post 'status': %s" % self._nodes(nodes))

    def cmd_update_pre(self, nodes):
        self.message("TestPlugin: Test pre 'update':  %s" % self._nodes(nodes))

    def cmd_update_post(self, results):
        self.message("TestPlugin: Test post 'update': %s" % self._results(results))

    def cmd_df_pre(self, nodes):
        self.message("TestPlugin: Test pre 'df':  %s" % self._nodes(nodes))

    def cmd_df_post(self, nodes):
        self.message("TestPlugin: Test post 'df': %s" % self._nodes(nodes))

    def cmd_diag_pre(self, nodes):
        self.message("TestPlugin: Test pre 'diag':  %s" % self._nodes(nodes))

    def cmd_diag_post(self, nodes):
        self.message("TestPlugin: Test post 'diag': %s" % self._nodes(nodes))

    def cmd_peerstatus_pre(self, nodes):
        self.message("TestPlugin: Test pre 'peerstatus':  %s" % self._nodes(nodes))

    def cmd_peerstatus_post(self, nodes):
        self.message("TestPlugin: Test post 'peerstatus': %s" % self._nodes(nodes))

    def cmd_netstats_pre(self, nodes):
        self.message("TestPlugin: Test pre 'netstats':  %s" % self._nodes(nodes))

    def cmd_netstats_post(self, nodes):
        self.message("TestPlugin: Test post 'netstats': %s" % self._nodes(nodes))

    def cmd_top_pre(self, nodes):
        self.message("TestPlugin: Test pre 'top':  %s" % self._nodes(nodes))

    def cmd_top_post(self, nodes):
        self.message("TestPlugin: Test post 'top': %s" % self._nodes(nodes))

    def cmd_cleanup_pre(self, nodes, all):
        self.message("TestPlugin: Test pre 'cleanup':  %s (%s)" % (self._nodes(nodes), all))

    def cmd_cleanup_post(self, nodes, all):
        self.message("TestPlugin: Test post 'cleanup': %s (%s)" % (self._nodes(nodes), all))

    def cmd_capstats_pre(self, nodes, interval):
        self.message("TestPlugin: Test pre 'capstats':  %s (%d)" % (self._nodes(nodes), interval))

    def cmd_capstats_post(self, nodes, interval):
        self.message("TestPlugin: Test post 'capstats':  %s (%d)" % (self._nodes(nodes), interval))

    def cmd_scripts_pre(self, nodes, check):
        self.message("TestPlugin: Test pre 'scripts':  %s (%s)" % (self._nodes(nodes), check))

    def cmd_scripts_post(self, nodes, check):
        self.message("TestPlugin: Test post 'scripts': %s (%s)" % (self._nodes(nodes), check))

    def cmd_print_pre(self, nodes, id):
        self.message("TestPlugin: Test pre 'print':  %s (%s)" % (self._nodes(nodes), id))

    def cmd_print_post(self, nodes, id):
        self.message("TestPlugin: Test post 'print': %s (%s)" % (self._nodes(nodes), id))

    def cmd_process_pre(self, trace, options, scripts):
        self.message("TestPlugin: Test pre 'process': %s %s -- %s" % (trace, options, scripts))
        return True

    def cmd_process_post(self, trace, options, scripts, success):
        self.message("TestPlugin: Test post 'process': %s %s -- %s -> %s" % (trace, options, scripts, success))






"""A Python type hierarchy representing Broker's data model.

Supported types are placed in a type hierarchy, with each type supporting
serialization to Broker's WebSocket wire format, unserialization from it, and
creation from "native" Python values, meaning of the closest typically used
Python type.

For reference, see: https://docs.zeek.org/projects/broker/en/current/web-socket.html

"""
import abc
import datetime
import enum
import ipaddress
import json
import re


class Type(abc.ABC):
    """Base class for types we can instantiate from or render to Broker's JSON
    data model. For details, see:
    https://docs.zeek.org/projects/broker/en/current/web-socket.html
    """

    def serialize(self, pretty=False):
        """Serializes the object to Broker-compatible wire data.

        pretty: When True, pretty-prints the resulting JSON.

        Returns: raw message data ready to transmit.
        """
        indent = 4 if pretty else None
        return json.dumps(self.to_broker(), indent=indent, sort_keys=True)

    def __eq__(self, other):
        """The default equality method for brokertypes.

        This implements member-by-member comparison based on the object's
        __dict__. The types complement this by each implementing their own
        __hash__() method.
        """
        if self.__class__ != other.__class__:
            return NotImplemented
        if len(self.__dict__) != len(other.__dict__):
            return False

        return self.__dict__ == other.__dict__

    def __repr__(self):
        return self.serialize()

    def __str__(self):
        return self.serialize(pretty=True)

    @classmethod
    def unserialize(cls, data):
        """Instantiates an object of this class from Broker wire data.

        This assumes the message content in JSON and first unserializes it into
        a Python data structure. It then calls from_broker() to instantiate an
        object of this class from it.

        data: raw wire WebSocket message content

        Returns: the resulting brokertype object.

        Raises: TypeError in case of invalid data. The exception's message
        provides details.
        """
        try:
            obj = json.loads(data)
        except json.JSONDecodeError as err:
            raise TypeError(
                f"cannot parse JSON data for {cls.__name__}: {err.msg} -- {data}"
            ) from err

        cls.check_broker_data(obj)

        try:
            # This may raise TypeError directly, which we pass on to the caller
            return cls.from_broker(obj)
        except (IndexError, KeyError, ValueError) as err:
            raise TypeError(f"invalid data for {cls.__name__}: {data}") from err

    @abc.abstractmethod
    def to_py(self):
        """Returns a Python-"native" rendering of the object.

        For most brokertypes this will be a native Python type (such as int or
        str), but for some types the closest thing to a natural rendering of the
        value in Python will be the object itself.

        Return: a Python value
        """
        return None

    @abc.abstractmethod
    def to_broker(self):
        """Returns a Broker-JSON-compatible Python data structure representing
        a value of this type.
        """
        return None

    @classmethod
    @abc.abstractmethod
    def check_broker_data(cls, data):
        """Checks the Broker data for compliance with the expected type.

        If you use unserialize() to obtain objects, you can ignore this
        method. The module invokes it under the hood.

        data: a Python data structure resulting from json.loads().

        Raises TypeError in case of problems.
        """

    @classmethod
    @abc.abstractmethod
    def from_broker(cls, data):
        """Returns an instance of the type given Broker's JSON data.

        This is a low-level method that you likely don't want to use. Consider
        unserialize() instead: it handles raw wire data unserialization,
        type-checking, and exception canonicalization.

        data: a JSON-unserialized Python data structure.

        Raises: type-specific exceptions resulting from value construction, such
        as TypeError, KeyError, or ValueError.
        """
        return None


# ---- Basic types -----------------------------------------------------


class DataType(Type):
    """Base class for data types known to Broker."""

    def __lt__(self, other):
        if not isinstance(other, DataType):
            raise TypeError(
                f"'<' comparison not supported between instances "
                f"of '{type(self).__name__}' and '{type(other).__name__}'"
            )
        # Supporting comparison accross data types allows us to sort the members
        # of a set or table keys. We simply compare the type names:
        if self.__class__ != other.__class__:
            return type(self).__name__ < type(other).__name__

        return NotImplemented

    @classmethod
    def check_broker_data(cls, data):
        if not isinstance(data, dict):
            raise TypeError("invalid data layout for Broker data: not an object")
        if "@data-type" not in data or "data" not in data:
            raise TypeError(
                "invalid data layout for Broker data: required keys missing"
            )


class NoneType(DataType):
    """Broker's representation of an absent value."""

    def __init__(self, _=None):
        # It helps to have a constructor that can be passed None explicitly, for
        # symmetry with other constructors below.
        pass

    def __lt__(self, other):
        res = super().__lt__(other)
        if res != NotImplemented:
            return res
        return False

    def __hash__(self):
        return hash(None)

    def to_py(self):
        return None

    def to_broker(self):
        return {
            "@data-type": "none",
            "data": {},
        }

    @classmethod
    def from_broker(cls, data):
        return NoneType()


class Boolean(DataType):
    def __init__(self, value):
        self._value = bool(value)

    def __lt__(self, other):
        res = super().__lt__(other)
        if res != NotImplemented:
            return res
        return self._value < other._value

    def __hash__(self):
        return hash(self._value)

    def to_py(self):
        return self._value

    def to_broker(self):
        return {
            "@data-type": "boolean",
            "data": self._value,
        }

    @classmethod
    def from_broker(cls, data):
        return Boolean(data["data"])


class Count(DataType):
    def __init__(self, value):
        self._value = int(value)
        if self._value < 0:
            raise ValueError("Count can only hold non-negative values")

    def __lt__(self, other):
        res = super().__lt__(other)
        if res != NotImplemented:
            return res
        return self._value < other._value

    def __hash__(self):
        return hash(self._value)

    def to_py(self):
        return self._value

    def to_broker(self):
        return {
            "@data-type": "count",
            "data": self._value,
        }

    @classmethod
    def from_broker(cls, data):
        return Count(data["data"])


class Integer(DataType):
    def __init__(self, value):
        self._value = int(value)

    def __lt__(self, other):
        res = super().__lt__(other)
        if res != NotImplemented:
            return res
        return self._value < other._value

    def __hash__(self):
        return hash(self._value)

    def to_py(self):
        return self._value

    def to_broker(self):
        return {
            "@data-type": "integer",
            "data": self._value,
        }

    @classmethod
    def from_broker(cls, data):
        return Integer(data["data"])


class Real(DataType):
    def __init__(self, value):
        self._value = float(value)

    def __lt__(self, other):
        res = super().__lt__(other)
        if res != NotImplemented:
            return res
        return self._value < other._value

    def __hash__(self):
        return hash(self._value)

    def to_py(self):
        return self._value

    def to_broker(self):
        return {
            "@data-type": "real",
            "data": self._value,
        }

    @classmethod
    def from_broker(cls, data):
        return Real(data["data"])


class Timespan(DataType):
    REGEX = re.compile(r"(\d+(\.\d+)?)(ns|ms|s|min|h|d)")

    class Unit(enum.Enum):
        """The time unit shorthands supported by Broker."""

        NS = "ns"
        MS = "ms"
        S = "s"
        MIN = "min"
        H = "h"
        D = "d"

    def __init__(self, value):
        if isinstance(value, datetime.timedelta):
            self._value = Timespan.timedelta_to_broker_timespan(value)
            self._td = value
        else:
            self._value = str(value)
            self._td = Timespan.broker_to_timedelta(self._value)

    def __eq__(self, other):
        # Make equality defined by the timedelta instances, not the
        # more variable string data (e.g. 1000ms == 1s):
        if self.__class__ != other.__class__:
            return False
        return self._td == other._td

    def __lt__(self, other):
        res = super().__lt__(other)
        if res != NotImplemented:
            return res
        return self._td < other._td

    def __hash__(self):
        return hash(self._value)

    def to_py(self):
        return self._td

    def to_broker(self):
        return {
            "@data-type": "timespan",
            "data": Timespan.timedelta_to_broker_timespan(self._td),
        }

    @classmethod
    def from_broker(cls, data):
        return Timespan(cls.broker_to_timedelta(data["data"]))

    @classmethod
    def broker_to_timedelta(cls, data):
        """Converts Broker-compatible timespan string into timedelta object."""
        mob = cls.REGEX.fullmatch(data)
        if mob is None:
            raise ValueError(f"'{data}' is not an acceptable Timespan value")

        counter = float(mob[1])
        unit = Timespan.Unit(mob[3])

        if unit == Timespan.Unit.NS:
            return datetime.timedelta(microseconds=counter / 1e3)
        if unit == Timespan.Unit.MS:
            return datetime.timedelta(milliseconds=counter)
        if unit == Timespan.Unit.S:
            return datetime.timedelta(seconds=counter)
        if unit == Timespan.Unit.MIN:
            return datetime.timedelta(minutes=counter)
        if unit == Timespan.Unit.H:
            return datetime.timedelta(hours=counter)
        if unit == Timespan.Unit.D:
            if counter % 7 == 0:
                return datetime.timedelta(weeks=counter / 7)
            return datetime.timedelta(days=counter)

        assert False, f"unhandled timespan unit '{unit}'"

    @classmethod
    def timedelta_to_broker_timespan(cls, tdelta):
        """Converts timedelta object to Broker-compatible timespan string."""
        # We use the smallest unit that's non-zero in the timespan (which has
        # only three relevant members: .microseconds, .seconds, and .days)
        # and map it to the closest Broker unit.

        def fmt(val, unit):
            # Don't say 10.0, say 10:
            val = int(val) if float(val).is_integer() else val
            return f"{val}{unit}"

        if tdelta.microseconds != 0:
            if tdelta.microseconds % 1000 == 0:
                return fmt(
                    tdelta.microseconds / 1e3
                    + tdelta.seconds * 1e3
                    + tdelta.days * 86400 * 1e3,
                    "ms",
                )
            # There are no microseconds in the Broker data model,
            # so go full plaid to nanoseconds.
            return fmt(
                tdelta.microseconds * 1e3
                + tdelta.seconds * 1e9
                + tdelta.days * 86400 * 1e9,
                "ns",
            )
        if tdelta.seconds != 0:
            if tdelta.seconds % 3600 == 0:
                return fmt(tdelta.seconds / 3600 + tdelta.days * 24, "h")
            if tdelta.seconds % 60 == 0:
                return fmt(tdelta.seconds / 60 + tdelta.days * 1440, "min")
            return fmt(tdelta.seconds + tdelta.days * 86400, "s")

        return fmt(tdelta.days, "d")


class Timestamp(DataType):
    def __init__(self, value):
        if isinstance(value, datetime.datetime):
            self._value = Timestamp.to_broker_iso8601(value)
            self._ts = value
        else:
            self._value = str(value)
            # Raise value error if not formatted acceptably
            self._ts = datetime.datetime.fromisoformat(self._value)

    def __eq__(self, other):
        # Make equality defined by the timestamp instances, not the
        # more variable ISO 8601 data:
        if self.__class__ != other.__class__:
            return False
        return self._ts == other._ts

    def __lt__(self, other):
        return self._ts < other._ts

    def __hash__(self):
        return hash(self._value)

    def to_py(self):
        return self._ts

    def to_broker(self):
        return {
            "@data-type": "timestamp",
            "data": Timestamp.to_broker_iso8601(self._ts),
        }

    @classmethod
    def from_broker(cls, data):
        return Timestamp(data["data"])

    @classmethod
    def to_broker_iso8601(cls, dtime):
        # The Broker docs say the timestamp looks like this:
        # "2022-04-10T07:00:00.000" -- meaning that given Python's
        # microseconds-granularity rendering we need to chop off the last
        # three digits:
        return dtime.isoformat(sep="T", timespec="microseconds")[:-3]


class String(DataType):
    def __init__(self, value):
        self._value = str(value)

    def __lt__(self, other):
        res = super().__lt__(other)
        if res != NotImplemented:
            return res
        return self._value < other._value

    def __hash__(self):
        return hash(self._value)

    def to_py(self):
        return self._value

    def to_broker(self):
        return {
            "@data-type": "string",
            "data": self._value,
        }

    @classmethod
    def from_broker(cls, data):
        return String(data["data"])


class Enum(DataType):
    def __init__(self, value):
        self._value = str(value)

    def __lt__(self, other):
        res = super().__lt__(other)
        if res != NotImplemented:
            return res
        return self._value < other._value

    def __hash__(self):
        return hash(self._value)

    def to_py(self):
        return self._value

    def to_broker(self):
        return {
            "@data-type": "enum-value",
            "data": self._value,
        }

    @classmethod
    def from_broker(cls, data):
        return Enum(data["data"])


class Address(DataType):
    def __init__(self, value):
        self._value = str(value)  # A str or ipaddress.IPv[46]Address
        # Throws a derivative of ValueError when not v4/v6 address:
        self._addr = ipaddress.ip_address(self._value)

    def __lt__(self, other):
        res = super().__lt__(other)
        if res != NotImplemented:
            return res
        return self._addr < other._addr

    def __hash__(self):
        return hash(self._value)

    def to_py(self):
        return self._addr

    def to_broker(self):
        return {
            "@data-type": "address",
            "data": self._value,
        }

    @classmethod
    def from_broker(cls, data):
        return Address(data["data"])


class Subnet(DataType):
    def __init__(self, value):
        self._value = str(value)  # A str or ipaddress.IPv[46]Network
        # Throws a derivative of ValueError when not v4/v6 network:
        self._subnet = ipaddress.ip_network(self._value)

    def __lt__(self, other):
        res = super().__lt__(other)
        if res != NotImplemented:
            return res
        return self._subnet < other._subnet

    def __hash__(self):
        return hash(self._value)

    def to_py(self):
        return self._subnet

    def to_broker(self):
        return {
            "@data-type": "subnet",
            "data": str(self._subnet),
        }

    @classmethod
    def from_broker(cls, data):
        return Subnet(data["data"])


class Port(DataType):
    class Proto(enum.Enum):
        UNKNOWN = "?"
        TCP = "tcp"
        UDP = "udp"
        ICMP = "icmp"

    def __init__(self, number, proto=Proto.TCP):
        self.number = int(number)
        self.proto = proto
        if not isinstance(proto, self.Proto):
            raise TypeError("Port constructor requires Proto enum")
        if self.number < 1 or self.number > 65535:
            raise ValueError(f"Port number '{self.number}' invalid")

    def __lt__(self, other):
        res = super().__lt__(other)
        if res != NotImplemented:
            return res
        order = ["?", "tcp", "udp", "icmp"]
        if order.index(self.proto.value) < order.index(other.proto.value):
            return True
        return self.number < other.number

    def __hash__(self):
        return hash((self.number, self.proto))

    def to_py(self):
        return self

    def to_broker(self):
        return {
            "@data-type": "port",
            "data": f"{self.number}/{self.proto.value}",
        }

    @classmethod
    def from_broker(cls, data):
        return Port(
            data["data"].split("/", 1)[0], Port.Proto(data["data"].split("/", 1)[1])
        )


class Vector(DataType):
    def __init__(self, elements=None):
        self._elements = elements or []
        if not isinstance(self._elements, tuple) and not isinstance(
            self._elements, list
        ):
            raise TypeError("Vector initialization requires tuple or list data")
        if not all(isinstance(elem, Type) for elem in self._elements):
            raise TypeError("Non-empty Vector construction requires brokertype values.")

    def __lt__(self, other):
        res = super().__lt__(other)
        if res != NotImplemented:
            return res
        for el1, el2 in zip(self._elements, other._elements):
            if el1 < el2:
                return True
        if len(self._elements) < len(other._elements):
            return True
        return False

    def __hash__(self):
        return hash(tuple(self._elements))

    def __iter__(self):
        return iter(self._elements)

    def __len__(self):
        return len(self._elements)

    def __getitem__(self, idx):
        return self._elements[idx]

    def to_py(self):
        return [elem.to_py() for elem in self._elements]

    def to_broker(self):
        return {
            "@data-type": "vector",
            "data": [elem.to_broker() for elem in self._elements],
        }

    @classmethod
    def from_broker(cls, data):
        return Vector([from_broker(elem) for elem in data["data"]])


class Set(DataType):
    def __init__(self, elements=None):
        self._elements = elements or set()
        if not isinstance(self._elements, set):
            raise TypeError("Set initialization requires set data")
        if not all(isinstance(elem, Type) for elem in self._elements):
            raise TypeError("Non-empty Set construction requires brokertype values.")

    def __lt__(self, other):
        res = super().__lt__(other)
        if res != NotImplemented:
            return res
        for el1, el2 in zip(sorted(self._elements), sorted(other._elements)):
            if el1 < el2:
                return True
        if len(self._elements) < len(other._elements):
            return True
        return False

    def __hash__(self):
        return hash(tuple(sorted(self._elements)))

    def __iter__(self):
        return iter(self._elements)

    def __len__(self):
        return len(self._elements)

    def __contains__(self, key):
        return key in self._elements

    def to_py(self):
        return set(elem.to_py() for elem in self._elements)

    def to_broker(self):
        return {
            "@data-type": "set",
            "data": [elem.to_broker() for elem in sorted(self._elements)],
        }

    @classmethod
    def from_broker(cls, data):
        return Set({from_broker(elem) for elem in data["data"]})


class Table(DataType):
    def __init__(self, elements=None):
        self._elements = elements or {}
        if not isinstance(self._elements, dict):
            raise TypeError("Table initialization requires dict data")
        keys_ok = all(isinstance(elem, Type) for elem in self._elements.keys())
        vals_ok = all(isinstance(elem, Type) for elem in self._elements.values())
        if not keys_ok or not vals_ok:
            raise TypeError("Non-empty Table construction requires brokertype values.")

    def __lt__(self, other):
        res = super().__lt__(other)
        if res != NotImplemented:
            return res
        for key1, key2 in zip(sorted(self._elements), sorted(other._elements)):
            if key1 < key2:
                return True
            if self._elements[key1] < other._elements[key2]:
                return True
        if len(self._elements) < len(other._elements):
            return True
        return False

    def __hash__(self):
        return hash((key, self._elements[key]) for key in sorted(self._elements))

    def __iter__(self):
        return iter(self._elements)

    def __len__(self):
        return len(self._elements)

    def __contains__(self, key):
        return key in self._elements

    def keys(self):
        return self._elements.keys()

    def values(self):
        return self._elements.values()

    def items(self):
        return self._elements.items()

    def to_py(self):
        res = {}
        for key, val in self._elements.items():
            res[key.to_py()] = val.to_py()
        return res

    def to_broker(self):
        return {
            "@data-type": "table",
            "data": [
                {"key": key.to_broker(), "value": self._elements[key].to_broker()}
                for key in sorted(self._elements)
            ],
        }

    @classmethod
    def from_broker(cls, data):
        return Table(
            {
                from_broker(elem["key"]): from_broker(elem["value"])
                for elem in data["data"]
            }
        )


# ---- Special types ---------------------------------------------------


class ZeekEvent(Vector):
    """Broker's event representation, as a vector of vectors.

    This specialization isn't an official type in Broker's hierarchy: there's no
    distinguishing @data-type for it. Zeek events are a specific interpretation
    of nested vectors.

    See Broker's websockets docs for an example:

    https://docs.zeek.org/projects/broker/en/current/web-socket.html#encoding-of-zeek-events
    """

    def __init__(self, name, *args):
        super().__init__()

        self.name = name.to_py() if isinstance(name, String) else str(name)
        self.args = list(args) or []  # list here is to avoid tuple/list type confusion

        for arg in self.args:
            if not isinstance(arg, Type):
                raise TypeError("ZeekEvent constructor requires brokertype arguments")

    def to_broker(self):
        return {
            "@data-type": "vector",
            "data": [
                {"@data-type": "count", "data": 1},
                {"@data-type": "count", "data": 1},
                {
                    "@data-type": "vector",
                    "data": [
                        String(self.name).to_broker(),
                        {
                            "@data-type": "vector",
                            "data": [arg.to_broker() for arg in self.args],
                        },
                    ],
                },
            ],
        }

    @classmethod
    def from_vector(cls, vec):
        """Special case for an existing Vector instance: recast as Zeek event."""
        if not isinstance(vec, Vector):
            raise TypeError("cannot convert non-vector to Zeek event")

        if (
            len(vec) < 3
            or not isinstance(vec[2], Vector)
            or len(vec[2]) < 2
            or not isinstance(vec[2][0], String)
            or not isinstance(vec[2][1], Vector)
        ):
            raise TypeError("invalid vector layout for Zeek event")

        name = vec[2][0].to_py()
        args = vec[2][1]

        # TODO: Extend to handle metadata

        return ZeekEvent(name, *args._elements)  # pylint: disable=protected-access

    @classmethod
    def from_broker(cls, data):
        name = data["data"][2]["data"][0]["data"]
        res = ZeekEvent(name)
        for argdata in data["data"][2]["data"][1]["data"]:
            res.args.append(from_broker(argdata))
        return res


# ---- Message types ---------------------------------------------------


class MessageType(Type):
    """Base class for Broker messages."""

    @classmethod
    def check_broker_data(cls, data):
        if not isinstance(data, dict):
            raise TypeError(
                f"invalid data layout for Broker {cls.__name__}: not an object"
            )
        if "type" not in data:
            raise TypeError(
                f"invalid data layout for Broker {cls.__name__}: required keys missing"
            )


class HandshakeMessage(MessageType):
    """The handshake message sent by the client.

    This is just a list of topics to subscribe to. Clients won't receive it.
    """

    def __init__(self, topics=None):
        self.topics = []

        if topics:
            if not isinstance(topics, tuple) and not isinstance(topics, list):
                raise TypeError("HandshakeMessage construction requires a topics list")
            for topic in topics:
                if isinstance(topic, str):
                    self.topics.append(topic)
                    continue
                if isinstance(topic, String):
                    self.topics.append(topic.to_py())
                    continue
                raise TypeError(
                    "topics for HandshakeMessage must be Python or "
                    "brokertype strings"
                )

    def to_py(self):
        return self

    def to_broker(self):
        return self.topics

    @classmethod
    def check_broker_data(cls, data):
        if not isinstance(data, tuple) and not isinstance(data, list):
            raise TypeError("invalid data layout for HandshakeMessage: not an object")

    @classmethod
    def from_broker(cls, data):
        return HandshakeMessage(data)


class HandshakeAckMessage(MessageType):
    """The ACK message returned to the client in response to the handshake.

    Clients won't need to send this.
    """

    def __init__(self, endpoint, version):
        self.endpoint = endpoint
        self.version = version

    def to_py(self):
        return self

    def to_broker(self):
        return {
            "type": "ack",
            "endpoint": self.endpoint,
            "version": self.version,
        }

    @classmethod
    def check_broker_data(cls, data):
        MessageType.check_broker_data(data)
        for key in ("type", "endpoint", "version"):
            if key not in data:
                raise TypeError(
                    f"invalid data layout for HandshakeAckMessage: "
                    f'required key "{key}" missing'
                )

    @classmethod
    def from_broker(cls, data):
        return HandshakeAckMessage(data["endpoint"], data["version"])


class DataMessage(MessageType):
    def __init__(self, topic, data):
        self.topic = topic
        self.data = data

    def to_py(self):
        return self

    def to_broker(self):
        bdata = self.data.to_broker()

        return {
            "type": "data-message",
            "topic": self.topic,
            "@data-type": bdata["@data-type"],
            "data": bdata["data"],
        }

    @classmethod
    def check_broker_data(cls, data):
        MessageType.check_broker_data(data)
        for key in ("type", "topic", "@data-type", "data"):
            if key not in data:
                raise TypeError(
                    f"invalid data layout for DataMessage: "
                    f'required key "{key}" missing'
                )

    @classmethod
    def from_broker(cls, data):
        return DataMessage(
            data["topic"],
            from_broker({"@data-type": data["@data-type"], "data": data["data"]}),
        )


class ErrorMessage(Type):
    def __init__(self, code, context):
        self.code = code  # A string representation of a Broker error code
        self.context = context

    def to_py(self):
        return self

    def to_broker(self):
        return {
            "type": "error",
            "code": self.code,
            "context": self.context,
        }

    @classmethod
    def check_broker_data(cls, data):
        MessageType.check_broker_data(data)
        for key in ("type", "code", "context"):
            if key not in data:
                raise TypeError(
                    f"invalid data layout for ErrorMessage: "
                    f'required key "{key}" missing'
                )

    @classmethod
    def from_broker(cls, data):
        return ErrorMessage(data["code"], data["context"])


# ---- Factory functions -----------------------------------------------

# This maps the types expressed in Broker's JSON representation to those
# implemented in this module.
_broker_typemap = {
    "none": NoneType,
    "address": Address,
    "boolean": Boolean,
    "count": Count,
    "enum-value": Enum,
    "integer": Integer,
    "port": Port,
    "real": Real,
    "set": Set,
    "string": String,
    "subnet": Subnet,
    "table": Table,
    "timespan": Timespan,
    "timestamp": Timestamp,
    "vector": Vector,
}

# This maps Broker's message types to ones implemented in this module.  A
# separate map, because Broker expresses the type information differently from
# the above.
_broker_messagemap = {
    "data-message": DataMessage,
    "error": ErrorMessage,
}


def unserialize(data):
    """A factory that instantiates a brokertype value from Broker wire data.

    This assumes the message content in JSON and first unserializes it into a
    Python data structure. It then calls from_python() to instantiate an object
    of the appropriate class from it.
    """
    try:
        obj = json.loads(data)
    except json.JSONDecodeError as err:
        raise TypeError(f"cannot parse JSON data: {err.msg} -- {data}") from err

    return from_broker(obj)


def from_broker(data):
    """A factory that turns Python-level data into brokertype instances.

    Consider using unserialize() instead, it starts from raw message data, and
    provides better error handling.

    data: a JSON-unserialized Python data structure.

    Returns: a brokerval instance

    Raises: TypeError in case of invalid input data.
    """
    if not isinstance(data, dict):
        raise TypeError("invalid data layout for Broker data: not an object")

    try:
        typ = _broker_messagemap[data["type"]]
        typ.check_broker_data(data)
        return typ.from_broker(data)
    except KeyError:
        pass

    try:
        typ = _broker_typemap[data["@data-type"]]
        typ.check_broker_data(data)
        return typ.from_broker(data)
    except KeyError as err:
        raise TypeError(f"unrecognized Broker type: {data}") from err


# Python types we can directly map to ones in this module, used by
# from_py(). This is imperfect since, for example, no non-negative integer type
# exists that maps to Count, but a generic factory adds convenience in many
# situations. Callers who need different mappings need to implement code that
# converts their data structures explicitly.
_python_typemap = {
    type(None): NoneType,
    bool: Boolean,
    datetime.timedelta: Timespan,
    datetime.datetime: Timestamp,
    dict: Table,
    float: Real,
    int: Integer,
    ipaddress.IPv4Address: Address,
    ipaddress.IPv6Address: Address,
    ipaddress.IPv4Network: Subnet,
    ipaddress.IPv6Network: Subnet,
    list: Vector,
    set: Set,
    str: String,
    tuple: Vector,
}


def from_py(data, typ=None, check_none=True):
    """Instantiates a brokertype object from the given Python data.

    Some Python types map naturally to Broker ones, such as bools and strs. For
    those, you can simply provide a value and the function will return the
    appropriate brokertype value. For some types this mapping isn't clear, and
    you need to specify the type explicitly. For composite types like
    sets or dicts the approach applies recursively to their member elements.

    When no type match is found, or the type conversion isn't feasible, this
    raises a TypeError. This can happen for types that don't have an immediate
    equivalent (e.g., Python has no unsigned integers).

    This function currently supports only types constructed from a single
    argument.

    data: a Python-"native" value, such as a str, int, or bool.

    typ (Type): if provided, the function attempts to instantiate an object of
        this type with the given data. By default, the function attempts type
        inference.

    check_none (bool): when True (the default), the function checks whether data
        is None, and shortcuts to returning a NoneType instance.

    Returns: a brokertype instance.

    Raises: TypeError in case problems arise in the type mapping or value
        construction.
    """
    if data is None and check_none:
        return NoneType()

    if typ is not None:
        if not issubclass(typ, Type):
            raise TypeError(f"not a brokertype: {typ.__name__}")
    else:
        try:
            typ = _python_typemap[type(data)]
        except KeyError as err:
            raise TypeError(
                f"cannot map Python type {type(data)} to Broker type"
            ) from err

    if typ == Table:
        res = Table()
        for key, val in data.items():
            res._elements[from_py(key)] = from_py(  # pylint: disable=protected-access
                val
            )
        return res

    if typ == Vector:
        res = Vector()
        for elem in data:
            res._elements.append(from_py(elem))  # pylint: disable=protected-access
        return res

    if typ == Set:
        res = Set()
        for elem in data:
            res._elements.add(from_py(elem))  # pylint: disable=protected-access
        return res

    # For others the constructors of the types in this module should naturally
    # work with the provided value.
    return typ(data)



# This file was automatically generated by SWIG (https://www.swig.org).
# Version 4.2.1
#
# Do not make changes to this file unless you know what you are doing - modify
# the SWIG interface file instead.

from sys import version_info as _swig_python_version_info
# Import the low-level C/C++ module
if __package__ or "." in __name__:
    from . import _SubnetTree
else:
    import _SubnetTree

try:
    import builtins as __builtin__
except ImportError:
    import __builtin__

def _swig_repr(self):
    try:
        strthis = "proxy of " + self.this.__repr__()
    except __builtin__.Exception:
        strthis = ""
    return "<%s.%s; %s >" % (self.__class__.__module__, self.__class__.__name__, strthis,)


def _swig_setattr_nondynamic_instance_variable(set):
    def set_instance_attr(self, name, value):
        if name == "this":
            set(self, name, value)
        elif name == "thisown":
            self.this.own(value)
        elif hasattr(self, name) and isinstance(getattr(type(self), name), property):
            set(self, name, value)
        else:
            raise AttributeError("You cannot add instance attributes to %s" % self)
    return set_instance_attr


def _swig_setattr_nondynamic_class_variable(set):
    def set_class_attr(cls, name, value):
        if hasattr(cls, name) and not isinstance(getattr(cls, name), property):
            set(cls, name, value)
        else:
            raise AttributeError("You cannot add class attributes to %s" % cls)
    return set_class_attr


def _swig_add_metaclass(metaclass):
    """Class decorator for adding a metaclass to a SWIG wrapped class - a slimmed down version of six.add_metaclass"""
    def wrapper(cls):
        return metaclass(cls.__name__, cls.__bases__, cls.__dict__.copy())
    return wrapper


class _SwigNonDynamicMeta(type):
    """Meta class to enforce nondynamic attributes (no new attributes) for a class"""
    __setattr__ = _swig_setattr_nondynamic_class_variable(type.__setattr__)


class inx_addr(object):
    thisown = property(lambda x: x.this.own(), lambda x, v: x.this.own(v), doc="The membership flag")
    __repr__ = _swig_repr
    sin = property(_SubnetTree.inx_addr_sin_get, _SubnetTree.inx_addr_sin_set)
    sin6 = property(_SubnetTree.inx_addr_sin6_get, _SubnetTree.inx_addr_sin6_set)

    def __init__(self):
        _SubnetTree.inx_addr_swiginit(self, _SubnetTree.new_inx_addr())
    __swig_destroy__ = _SubnetTree.delete_inx_addr

# Register inx_addr in _SubnetTree:
_SubnetTree.inx_addr_swigregister(inx_addr)
class SubnetTree(object):
    thisown = property(lambda x: x.this.own(), lambda x, v: x.this.own(v), doc="The membership flag")
    __repr__ = _swig_repr

    def __init__(self, binary_lookup_mode=False):
        _SubnetTree.SubnetTree_swiginit(self, _SubnetTree.new_SubnetTree(binary_lookup_mode))
    __swig_destroy__ = _SubnetTree.delete_SubnetTree

    def insert(self, *args):
        return _SubnetTree.SubnetTree_insert(self, *args)

    def remove(self, *args):
        return _SubnetTree.SubnetTree_remove(self, *args)

    def lookup(self, *args):
        return _SubnetTree.SubnetTree_lookup(self, *args)

    def prefixes(self, ipv4_native=False, with_len=True):
        return _SubnetTree.SubnetTree_prefixes(self, ipv4_native, with_len)

    def get_binary_lookup_mode(self):
        return _SubnetTree.SubnetTree_get_binary_lookup_mode(self)

    def set_binary_lookup_mode(self, binary_lookup_mode=True):
        return _SubnetTree.SubnetTree_set_binary_lookup_mode(self, binary_lookup_mode)

    def __contains__(self, *args):
        return _SubnetTree.SubnetTree___contains__(self, *args)

    def __getitem__(self, cidr):
        return _SubnetTree.SubnetTree___getitem__(self, cidr)

    def __setitem__(self, cidr, data):
        return _SubnetTree.SubnetTree___setitem__(self, cidr, data)

    def __delitem__(self, cidr):
        return _SubnetTree.SubnetTree___delitem__(self, cidr)

# Register SubnetTree in _SubnetTree:
_SubnetTree.SubnetTree_swigregister(SubnetTree)




#
# A simple example plugin implementing a command "ps.zeek" that shows all Zeek
# processes currently running on any of the configured nodes, including  those
# not controlled by this zeekctl. The latter are marked with "(-)", while "our"
# processes get a "(+)".

import ZeekControl.plugin
import ZeekControl.cmdresult

class PsZeek(ZeekControl.plugin.Plugin):
    def __init__(self):
        super(PsZeek, self).__init__(apiversion=1)

    def name(self):
        return "ps"

    def pluginVersion(self):
        return 1

    def commands(self):
        return [
            ("zeek", "[<nodes>]", "Show Zeek processes on nodes' systems"),
        ]

    def cmd_custom(self, cmd, args, cmdout):
        results = ZeekControl.cmdresult.CmdResult()

        assert(cmd == "zeek") # Can't be anything else.

        # Get the nodes the user wants.
        if args:
            nodes, notnodes = self.parseNodes(args)
            for n in notnodes:
                cmdout.error("unknown node '%s'" % n)
        else:
            nodes = self.nodes()

        if not nodes:
            cmdout.error("No nodes given.")
            results.ok = False
            return results

        # Get one node for every host running at least one of the nodes. Also
        # records each hosts Zeek PIDs.
        host_nodes = {}
        pids = {}

        for n in nodes:
            pid = n.getPID()
            if pid:
                p = pids.setdefault(n.host, set())
                p.add(pid)

            host_nodes[n.host] = n

        # Build commands to execute.

        # The grep command grabs the header line and all lines ending in "zeek"
        # (and ignores "run-zeek" that may appear in the output of ps).
        cmd = "POSIXLY_CORRECT=1 ps axco user,pid,ppid,%cpu,%mem,vsz,rss,tt,state,start,time,command | grep -e PID -e '[^-]zeek$'"
        cmds = [(n, cmd) for n in host_nodes.values()]
        cmds.sort(key=lambda n: n[0].name)

        # Run them in parallel and print output.

        first_node = True

        for (n, success, output) in self.executeParallel(cmds):
            outlines = output.splitlines()
            # Remove stderr output (if any)
            while outlines and not outlines[0].startswith("USER"):
                outlines = outlines[1:]

            # Print the header line.
            if first_node and outlines:
                cmdout.info("        %s" % outlines[0])

            if success:
                cmdout.info(">>> %s" % n.host)
            else:
                cmdout.error(">>> %s failed" % n.host)
                results.ok = False

            if not outlines:
                continue

            for line in outlines[1:]:

                m = line.split()
                try:
                    pid, ppid = int(m[1]), int(m[2])
                except IndexError:
                    cmdout.error("unexpected output from ps command: %s" % line)
                    results.ok = False
                    continue
                except ValueError as err:
                    cmdout.error("%s" % err)
                    results.ok = False
                    continue
                try:
                    known = (pid in pids[n.host] or ppid in pids[n.host])
                except KeyError:
                    known = False

                if known:
                    cmdout.info("   (+) %s" % line.strip())
                else:
                    cmdout.info("   (-) %s" % line.strip())

            first_node = False

        return results




"""Infrastructure for typed Zeek events."""
from .logs import LOG

from .brokertypes import (
    Boolean,
    Set,
    String,
    Type,
    Vector,
    ZeekEvent,
    from_py,
)

from .types import (
    SerializableZeekType,
)


class Event(SerializableZeekType):
    NAME = None  # Name of the event, e.g. "Management::Controller::API::deploy_request"
    ARG_NAMES = []  # Names of the arguments, e.g. "reqid"
    ARG_TYPES = []  # Types in Python, e.g. str

    def __init__(self, *args):
        """Creates a Zeek event object.

        This expects the number of arguments contextualized above. The event
        name is not required since it's defined implicitly via the event class
        receiving the arguments.

        Raises:
            TypeError: when the given arguments, or number of arguments, don't
                match the expected ARG_TYPES or their number.
        """
        if len(args) != len(self.ARG_NAMES):
            raise TypeError(
                f"event argument length mismatch: have {len(args)}, expected {len(self.ARG_NAMES)}"
            )

        self.args = []

        for idx, arg in enumerate(args):
            # If the argument's type matches the required Broker type, we're done.
            if isinstance(arg, self.ARG_TYPES[idx]):
                self.args.append(arg)
                continue

            try:
                # When creating an event it can be convenient for the caller to
                # pass Python-native types. See if we can create brokertypes
                # types from them, to match the types actually specified when we
                # created the event classes.
                maybe_arg = from_py(arg)
            except TypeError as err:
                raise TypeError(
                    f"event argument type mismatch: argument "
                    f"{idx+1} is {type(arg)}, {err}"
                ) from err

            # Again: if we now have a type match, we're done.
            if isinstance(maybe_arg, self.ARG_TYPES[idx]):
                self.args.append(maybe_arg)
                continue

            raise TypeError(
                f"event argument type mismatch: argument "
                f"{idx+1} is {type(arg)}, should be {self.ARG_TYPES[idx]}"
            )

    def __getattr__(self, name):
        """Allow attribute-like access to event arguments."""
        try:
            idx = self.ARG_NAMES.index(name)
            return self.args[idx]
        except ValueError as err:
            raise AttributeError(
                f'event type {self.NAME} has no "{name}" argument'
            ) from err

    def __str__(self):
        # A list of pairs (argument name, typename)
        zeek_style_args = zip(self.ARG_NAMES, [str(type(arg)) for arg in self.args])
        # That list, with each item now a string "<name>: <typename"
        zeek_style_arg_strings = [": ".join(arg) for arg in zeek_style_args]
        # A Zeek-looking event signature
        return self.NAME + "(" + ", ".join(zeek_style_arg_strings) + ")"

    def to_brokertype(self):
        return ZeekEvent(self.NAME, *self.args)

    @classmethod
    def from_brokertype(cls, data):
        # Verify that data is an event
        return Registry.make_event(data.name, data.args)


class Registry:
    """Functionality for event types and to instantiate typed events from data."""

    # Map from Zeek-level event names to Event classes. The make_event()
    # function uses this map to instantiate the right event class from
    # received Broker data.
    EVENT_TYPES = {}

    @staticmethod
    def make_event_class(name, arg_names, arg_types):
        """Factory function to generate a Zeek event class.

        Given an event name, event arguments, and corresponding argument types,
        the function generates a new Event class, registers it, and returns it.
        """
        res = type(name, (Event,), {})

        if len(arg_names) != len(arg_types):
            raise TypeError(
                f"error creating event type {name}: number of event "
                f"argument names and types must match ({len(arg_names)}/{len(arg_types)})"
            )

        for idx, typ in enumerate(arg_types):
            if not issubclass(typ, Type):
                raise TypeError(
                    f"event type creation error: argument {idx+1}, "
                    f'"{arg_names[idx]}", is not a brokertype class'
                )
        res.NAME = name
        res.ARG_NAMES = arg_names
        res.ARG_TYPES = arg_types

        # Register the new event type
        Registry.EVENT_TYPES[name] = res

        return res

    @staticmethod
    def make_event(name, *args):
        """This method allows constructing an Event instance from its name."""
        if name not in Registry.EVENT_TYPES:
            LOG.warning('received unexpected event "%s", skipping', name)
            return None

        LOG.debug('received event "%s"', name)
        return Registry.EVENT_TYPES[name](*args)


# Any Zeek object/record that's an event argument gets represented as a
# tuple below, reflecting Broker's representation thereof.

DeployRequest = Registry.make_event_class(
    "Management::Controller::API::deploy_request", ("reqid",), (String,)
)

DeployResponse = Registry.make_event_class(
    "Management::Controller::API::deploy_response",
    ("reqid", "results"),
    (String, Vector),
)

GetConfigurationRequest = Registry.make_event_class(
    "Management::Controller::API::get_configuration_request",
    ("reqid", "deployed"),
    (String, Boolean),
)

GetConfigurationResponse = Registry.make_event_class(
    "Management::Controller::API::get_configuration_response",
    ("reqid", "result"),
    (String, Vector),
)

GetIdValueRequest = Registry.make_event_class(
    "Management::Controller::API::get_id_value_request",
    ("reqid", "id", "nodes"),
    (String, String, Set),
)

GetIdValueResponse = Registry.make_event_class(
    "Management::Controller::API::get_id_value_response",
    ("reqid", "results"),
    (String, Vector),
)

GetInstancesRequest = Registry.make_event_class(
    "Management::Controller::API::get_instances_request", ("reqid",), (String,)
)

GetInstancesResponse = Registry.make_event_class(
    "Management::Controller::API::get_instances_response",
    ("reqid", "result"),
    (String, Vector),
)

GetNodesRequest = Registry.make_event_class(
    "Management::Controller::API::get_nodes_request", ("reqid",), (String,)
)

GetNodesResponse = Registry.make_event_class(
    "Management::Controller::API::get_nodes_response",
    ("reqid", "results"),
    (String, Vector),
)

RestartRequest = Registry.make_event_class(
    "Management::Controller::API::restart_request", ("reqid", "nodes"), (String, Set)
)

RestartResponse = Registry.make_event_class(
    "Management::Controller::API::restart_response",
    ("reqid", "results"),
    (String, Vector),
)

StageConfigurationRequest = Registry.make_event_class(
    "Management::Controller::API::stage_configuration_request",
    ("reqid", "config"),
    (String, Vector),
)

StageConfigurationResponse = Registry.make_event_class(
    "Management::Controller::API::stage_configuration_response",
    ("reqid", "results"),
    (String, Vector),
)

TestNoopRequest = Registry.make_event_class(
    "Management::Controller::API::test_noop_request", ("reqid",), (String,)
)

TestTimeoutRequest = Registry.make_event_class(
    "Management::Controller::API::test_timeout_request",
    ("reqid", "with_state"),
    (String, Boolean),
)

TestTimeoutResponse = Registry.make_event_class(
    "Management::Controller::API::test_timeout_response",
    ("reqid", "result"),
    (String, Vector),
)



import uuid


def make_uuid(prefix=""):
    """Returns a prefixable  UUID string.

    The client uses this for event request/response identifiers.

    Args:
        prefix (string): prefix string to prepend to the UUID.
    """
    return prefix + str(uuid.uuid1())



import os

# The Broker topic prefix for communicating with the controller
CONTROLLER_TOPIC = "zeek/management/controller"

# The default location zeek-client considers for a config file
CONFIG_FILE = os.getenv("ZEEK_CLIENT_CONFIG_FILE") or "/etc/zeek/zeek-client.cfg"



import os
import time

from ZeekControl import config

lockCount = 0

# Return: 0 if no lock, >0 for PID of lock, or -1 on error
def _break_lock(cmdout):
    from ZeekControl import execute

    try:
        # Check whether lock is stale.
        with open(config.Config.lockfile, "r") as f:
            pid = f.readline().strip()

    except (OSError, IOError) as err:
        cmdout.error("failed to read lock file: %s" % err)
        return -1

    success, output = execute.run_localcmd("%s %s" % (os.path.join(config.Config.helperdir, "check-pid"), pid))
    if success and output.strip() == "running":
        # Process still exists.
        try:
            return int(pid)
        except ValueError:
            return -1

    cmdout.info("removing stale lock")
    try:
        # Break lock.
        os.unlink(config.Config.lockfile)
    except (OSError, IOError) as err:
        cmdout.error("failed to remove lock file: %s" % err)
        return -1

    return 0

# Return: 0 if lock is acquired, or if failed to acquire lock return >0 for
# PID of lock, or -1 on error
def _acquire_lock(cmdout):
    lockpid = -1
    pid = str(os.getpid())
    tmpfile = config.Config.lockfile + "." + pid

    lockdir = os.path.dirname(config.Config.lockfile)
    if not os.path.exists(lockdir):
        cmdout.info("creating directory for lock file: %s" % lockdir)
        os.makedirs(lockdir)

    try:
        try:
            # This should be NFS-safe.
            with open(tmpfile, "w") as f:
                f.write("%s\n" % pid)

            n = os.stat(tmpfile)[3]
            os.link(tmpfile, config.Config.lockfile)
            m = os.stat(tmpfile)[3]

            if n == m-1:
                return 0

            # File is locked.
            lockpid = _break_lock(cmdout)
            if lockpid == 0:
                return _acquire_lock(cmdout)

        except OSError:
            # File is already locked.
            lockpid = _break_lock(cmdout)
            if lockpid == 0:
                return _acquire_lock(cmdout)

        except IOError as e:
            cmdout.error("cannot acquire lock: %s" % e)
            return lockpid

    finally:
        try:
            os.unlink(tmpfile)
        except (OSError, IOError):
            pass

    return lockpid

def _release_lock(cmdout):
    try:
        os.unlink(config.Config.lockfile)
    except OSError as e:
        cmdout.error("cannot remove lock file: %s" % e)

def lock(cmdout, showwait=True):
    global lockCount

    if lockCount > 0:
        # Already locked.
        lockCount += 1
        return True

    lockpid = _acquire_lock(cmdout)
    if lockpid < 0:
        return False

    if lockpid:
        if showwait:
            cmdout.info("waiting for lock (owned by PID %d) ..." % lockpid)

        count = 0
        while _acquire_lock(cmdout) != 0:
            time.sleep(1)

            count += 1
            if count > 30:
                return False

    lockCount = 1
    return True

def unlock(cmdout):
    global lockCount

    if lockCount == 0:
        cmdout.error("mismatched lock/unlock")
        return

    if lockCount > 1:
        # Still locked.
        lockCount -= 1
        return

    _release_lock(cmdout)

    lockCount = 0
